from typing import *

import numpy as np
from numpy import ndarray


# P-Norm
def p_norm_numpy(values: ndarray, targets: ndarray, p=2, axis: Union[int, Sequence[int]] = None) -> ndarray:
    return np.sum(np.abs(values - targets) ** p, axis=axis) ** (1 / p)


# Mean Absolute Error
def mae_numpy(value: ndarray, targets: ndarray, axis: Union[int, Sequence[int]] = None) -> ndarray:
    return np.mean(np.abs(value - targets), axis=axis)


# Mean Squared Error
def mse_numpy(values: ndarray, targets: ndarray, axis: Union[int, Sequence[int]] = None) -> ndarray:
    return np.mean(np.square(values - targets), axis=axis)


# Root Mean Squared Error
def rmse_numpy(values: ndarray, targets: ndarray, axis: Union[int, Sequence[int]] = None) -> ndarray:
    return np.sqrt(np.mean(np.square(values - targets), axis=axis))


# Root Mean Squared Logarithmic Error
def rmlse_numpy(values: ndarray, targets: ndarray, axis: Union[int, Sequence[int]] = None) -> ndarray:
    return np.sqrt(np.mean(np.square(np.log(values + 1) - np.log(targets + 1)), axis=axis))


# Logarithmic Mean Squared Error
def lmse_numpy(values: ndarray, targets: ndarray, axis: Union[int, Sequence[int]] = None) -> ndarray:
    return np.log(np.mean(np.square(values - targets), axis=axis))


# Cosine
def cosine_numpy(values: ndarray, targets: ndarray, axis: Union[int, Sequence[int]] = None) -> ndarray:
    if np.all(np.equal(values, 0.0)) or np.all(np.equal(targets, 0.0)):
        return 1.0
    return np.sum(values * targets, axis=axis) / (
            np.sqrt(np.sum(np.square(values), axis=axis)) * np.sqrt(np.sum(np.square(targets), axis=axis)))


# Angular
def angular_numpy(values: ndarray, targets: ndarray, axis: Union[int, Sequence[int]] = None) -> ndarray:
    return 2 * np.arccos(cosine_numpy(values, targets, axis=axis)) / np.pi


def loss_numpy(name: str) -> Callable[[ndarray, ndarray, Optional[int]], Union[float, ndarray]]:
    if name[-4:] == "norm":
        p = int(name[:-5])
        return lambda x, y, axis=None: p_norm_numpy(x, y, p, axis=axis)
    elif name == "mae":
        return lambda x, y, axis=None: mae_numpy(x, y, axis=axis)
    elif name == "mse":
        return lambda x, y, axis=None: mse_numpy(x, y, axis=axis)
    elif name == "rmse":
        return lambda x, y, axis=None: rmse_numpy(x, y, axis=axis)
    elif name == "rmlse":
        return lambda x, y, axis=None: rmlse_numpy(x, y, axis=axis)
    elif name == "lmse":
        return lambda x, y, axis=None: lmse_numpy(x, y, axis=axis)
    elif name == "cosine":
        return lambda x, y, axis=None: cosine_numpy(x, y, axis=axis)
    elif name == "angular":
        return lambda x, y, axis=None: angular_numpy(x, y, axis=axis)
    else:
        raise NotImplementedError(name)



__all__ = ["DatasetLoader"]

import logging
import os
import typing

import openpyxl
import xlrd
from PySide6 import QtCore, QtWidgets

from ..models import Dataset
from ..io.load import check_layout, load_dataset, get_file_type, ReadFileType


class DatasetLoader(QtWidgets.QDialog):
    logger = logging.getLogger("QGrain.DatasetLoader")
    dataset_loaded = QtCore.Signal(Dataset)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.setWindowTitle(self.tr("Dataset Loader"))
        self.main_layout = QtWidgets.QGridLayout(self)
        self.filename_display = QtWidgets.QLabel(self.tr("Filename Unknown"))
        self.main_layout.addWidget(self.filename_display, 0, 0)
        self.select_button = QtWidgets.QPushButton(self.tr("Select"))
        self.select_button.clicked.connect(self.on_select_clicked)
        self.main_layout.addWidget(self.select_button, 0, 1)
        self.sheet_label = QtWidgets.QLabel(self.tr("Sheet Name"))
        self.sheet_combo_box = QtWidgets.QComboBox()
        self.sheet_combo_box.addItem(self.tr("Empty"))
        self.main_layout.addWidget(self.sheet_label, 1, 0)
        self.main_layout.addWidget(self.sheet_combo_box, 1, 1)
        self.class_row_label = QtWidgets.QLabel(self.tr("Row With Grain Size Classes"))
        self.class_row_input = QtWidgets.QSpinBox()
        self.class_row_input.setRange(1, 999)
        self.main_layout.addWidget(self.class_row_label, 2, 0)
        self.main_layout.addWidget(self.class_row_input, 2, 1)
        self.name_column_label = QtWidgets.QLabel(self.tr("Column With Sample Names"))
        self.name_column_input = QtWidgets.QSpinBox()
        self.name_column_input.setRange(1, 999)
        self.main_layout.addWidget(self.name_column_label, 3, 0)
        self.main_layout.addWidget(self.name_column_input, 3, 1)
        self.start_row_label = QtWidgets.QLabel(self.tr("Start Row of Distributions"))
        self.start_row_input = QtWidgets.QSpinBox()
        self.start_row_input.setRange(2, 999999)
        self.main_layout.addWidget(self.start_row_label, 4, 0)
        self.main_layout.addWidget(self.start_row_input, 4, 1)
        self.start_column_label = QtWidgets.QLabel(self.tr("Start Column of Distributions"))
        self.start_column_input = QtWidgets.QSpinBox()
        self.start_column_input.setRange(2, 999999)
        self.main_layout.addWidget(self.start_column_label, 5, 0)
        self.main_layout.addWidget(self.start_column_input, 5, 1)
        self.try_load_button = QtWidgets.QPushButton(self.tr("Try Load"))
        self.try_load_button.clicked.connect(self.on_try_load_clicked)
        self.try_load_button.setEnabled(False)
        self.main_layout.addWidget(self.try_load_button, 6, 0, 1, 2)
        self.normal_msg = QtWidgets.QMessageBox(self)
        self.file_dialog = QtWidgets.QFileDialog(parent=self)
        self._filename = ""

    @property
    def sheet_index(self) -> int:
        sheet_index = self.sheet_combo_box.currentIndex()
        return sheet_index

    @property
    def sheet_name(self) -> str:
        sheet_name = self.sheet_combo_box.currentText()
        return sheet_name

    @property
    def data_layout(self) -> typing.Optional[typing.Dict[str, int]]:
        class_row = self.class_row_input.value() - 1
        name_col = self.name_column_input.value() - 1
        start_row = self.start_row_input.value() - 1
        start_col = self.start_column_input.value() - 1
        layout = dict(class_row=class_row, name_col=name_col, start_row=start_row, start_col=start_col)
        return layout

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    def on_select_clicked(self):
        filename, _ = self.file_dialog.getOpenFileName(
            self, self.tr("Select a file which stores the grain size dataset"), ".",
            "Excel (*.xlsx);;97-2003 Excel (*.xls);;CSV (*.csv)")
        if filename is None or filename == "":
            return
        self._filename = filename
        file_type = get_file_type(filename)
        if file_type == ReadFileType.CSV:
            sheet_names = [os.path.basename(filename)]
        elif file_type == ReadFileType.XLS:
            workbook = xlrd.open_workbook(filename)
            sheet_names = workbook.sheet_names()
            workbook.release_resources()
        elif file_type == ReadFileType.XLSX:
            workbook = openpyxl.load_workbook(filename, read_only=True, data_only=True)
            sheet_names = workbook.sheetnames
            workbook.close()
        else:
            raise NotImplementedError(file_type)
        self.filename_display.setText(os.path.basename(filename))
        self.sheet_combo_box.clear()
        self.sheet_combo_box.addItems(sheet_names)
        self.try_load_button.setEnabled(True)

    def on_try_load_clicked(self):
        try:
            check_layout(**self.data_layout)
        except ValueError:
            self.logger.error("The current layout setting is invalid.")
            self.show_error(self.tr("The current layout setting is invalid."))
            return

        progress_dialog = QtWidgets.QProgressDialog(
            self.tr("Loading the grain size distributions..."), self.tr("Cancel"),
            0, 100, self)
        progress_dialog.setWindowTitle("QGrain")
        progress_dialog.setWindowModality(QtCore.Qt.WindowModal)

        def callback(progress: float):
            if progress_dialog.wasCanceled():
                raise StopIteration()
            progress_dialog.setValue(int(progress * 100))
            QtCore.QCoreApplication.processEvents()
        try:
            result = load_dataset(self._filename, sheet_index=self.sheet_index,  **self.data_layout,
                                  progress_callback=callback, logger=self.logger)
            if result is None:
                self.show_error(self.tr("Can not load the grain size dataset from this file. "
                                        "Please check the logs for more details."))
            else:
                progress_dialog.setValue(100)
                self.dataset_loaded.emit(result)
                self.logger.info("Good job! The grain size dataset has been loaded from this file, "
                                 "and has been emitted to other widgets.")
                self.hide()
        except StopIteration:
            self.logger.info("The loading task was canceled.")
        finally:
            progress_dialog.close()

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("Dataset Loader"))
        if self._filename is None:
            self.filename_display.setText(self.tr("Filename Unspecified"))
            self.sheet_combo_box.setItemText(0, self.tr("Empty"))
        self.select_button.setText(self.tr("Select"))
        self.sheet_label.setText(self.tr("Sheet Name"))
        self.class_row_label.setText(self.tr("Row With Grain Size Classes"))
        self.name_column_label.setText(self.tr("Column With Sample Names"))
        self.start_row_label.setText(self.tr("Start Row of Distributions"))
        self.start_column_label.setText(self.tr("Start Column of Distributions"))
        self.try_load_button.setText(self.tr("Try Load"))



__all__ = ["SSUSettings"]

from typing import *

from PySide6 import QtCore, QtWidgets


class SSUSettings(QtWidgets.QDialog):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setWindowTitle(self.tr("SSU Settings"))
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.loss_label = QtWidgets.QLabel(self.tr("Loss Function"))
        self.loss_label.setToolTip(self.tr(
            "The function to calculate the difference between prediction and observation."))
        self.loss_combo_box = QtWidgets.QComboBox()
        for key, name in self.supported_losses:
            self.loss_combo_box.addItem(name)
        self.loss_combo_box.setCurrentIndex(8)
        self.main_layout.addWidget(self.loss_label, 0, 0)
        self.main_layout.addWidget(self.loss_combo_box, 0, 1)
        self.optimizer_label = QtWidgets.QLabel(self.tr("Optimizer"))
        self.optimizer_label.setToolTip(self.tr("The optimizer to find the minimum of loss function."))
        self.optimizer_combo_box = QtWidgets.QComboBox()
        for key, name in self.supported_optimizers:
            self.optimizer_combo_box.addItem(name)
        self.optimizer_combo_box.setCurrentIndex(6)
        self.main_layout.addWidget(self.optimizer_label, 1, 0)
        self.main_layout.addWidget(self.optimizer_combo_box, 1, 1)
        self.try_global_checkbox = QtWidgets.QCheckBox(self.tr("Global Optimization"))
        self.try_global_checkbox.setChecked(False)
        self.try_global_checkbox.setToolTip(self.tr("Try global optimization or not."))
        self.try_global_checkbox.clicked.connect(self.on_try_global_changed)
        self.main_layout.addWidget(self.try_global_checkbox, 2, 0, 1, 2)
        self.global_max_niter_label = QtWidgets.QLabel(self.tr("Maximum Number of Iterations"))
        self.global_max_niter_label.setToolTip(self.tr("Maximum number of iterations of global optimization."))
        self.global_max_niter_input = QtWidgets.QSpinBox()
        self.global_max_niter_input.setRange(0, 10000)
        self.global_max_niter_input.setValue(100)
        self.main_layout.addWidget(self.global_max_niter_label, 3, 0)
        self.main_layout.addWidget(self.global_max_niter_input, 3, 1)
        self.global_niter_success_label = QtWidgets.QLabel(self.tr("Success Number of Iterations"))
        self.global_niter_success_label.setToolTip(self.tr(
            "The number of iterations that reaching the same minimum value."))
        self.global_niter_success_input = QtWidgets.QSpinBox()
        self.global_niter_success_input.setRange(1, 100)
        self.global_niter_success_input.setValue(5)
        self.main_layout.addWidget(self.global_niter_success_label, 4, 0)
        self.main_layout.addWidget(self.global_niter_success_input, 4, 1)
        self.global_step_size_label = QtWidgets.QLabel(self.tr("Step Size"))
        self.global_step_size_label.setToolTip(self.tr("The step size of searching the global minimum."))
        self.global_step_size_input = QtWidgets.QDoubleSpinBox()
        self.global_step_size_input.setRange(0.01, 10)
        self.global_step_size_input.setValue(0.1)
        self.main_layout.addWidget(self.global_step_size_label, 5, 0)
        self.main_layout.addWidget(self.global_step_size_input, 5, 1)
        self.optimizer_max_niter_label = QtWidgets.QLabel(self.tr("Maximum Number of Iterations of Optimizer"))
        self.optimizer_max_niter_label.setToolTip(self.tr(
            "Maximum number of iterations of the optimizer in global and local optimization."))
        self.optimizer_max_niter_input = QtWidgets.QSpinBox()
        self.optimizer_max_niter_input.setRange(0, 100000)
        self.optimizer_max_niter_input.setValue(500)
        self.main_layout.addWidget(self.optimizer_max_niter_label, 6, 0)
        self.main_layout.addWidget(self.optimizer_max_niter_input, 6, 1)
        self.need_history_checkbox = QtWidgets.QCheckBox(self.tr("Need History"))
        self.need_history_checkbox.setChecked(True)
        self.need_history_checkbox.setToolTip(self.tr("Record the variation history of parameters or not."))
        self.main_layout.addWidget(self.need_history_checkbox, 7, 0, 1, 2)

    @property
    def supported_losses(self) -> Sequence[Tuple[str, str]]:
        losses = (("1-norm", self.tr("1 Norm")),
                  ("2-norm", self.tr("2 Norm")),
                  ("3-norm", self.tr("3 Norm")),
                  ("4-norm", self.tr("4 Norm")),
                  ("mae", self.tr("MAE")),
                  ("mse", self.tr("MSE")),
                  ("rmse", self.tr("RMSE")),
                  ("rmlse", self.tr("RMLSE")),
                  ("lmse", self.tr("LMSE")),
                  ("cosine", self.tr("Cosine")),
                  ("angular", self.tr("Angular")))
        return losses

    @property
    def supported_optimizers(self) -> Sequence[Tuple[str, str]]:
        optimizers = (("Nelder-Mead", self.tr("Nelder-Mead")),
                      ("Powell", self.tr("Powell")),
                      ("CG", self.tr("CG")),
                      ("BFGS", self.tr("BFGS")),
                      ("L-BFGS-B", self.tr("L-BFGS-B")),
                      ("TNC", self.tr("TNC")),
                      ("SLSQP", self.tr("SLSQP")))
        return optimizers

    @property
    def loss(self) -> str:
        return self.supported_losses[self.loss_combo_box.currentIndex()][0]

    @property
    def optimizer(self) -> str:
        return self.supported_optimizers[self.optimizer_combo_box.currentIndex()][0]

    @property
    def settings(self) -> Dict:
        s = dict(
            loss=self.loss,
            optimizer=self.optimizer,
            try_global=self.try_global_checkbox.isChecked(),
            global_max_niter=self.global_max_niter_input.value(),
            global_niter_success=self.global_niter_success_input.value(),
            global_step_size=self.global_step_size_input.value(),
            optimizer_max_niter=self.optimizer_max_niter_input.value(),
            need_history=self.need_history_checkbox.isChecked())
        return s

    @settings.setter
    def settings(self, s: Dict[str, Any]):
        loss_map = {key: i for i, (key, name) in enumerate(self.supported_losses)}
        assert s["loss"] in loss_map
        optimizer_map = {key: i for i, (key, name) in enumerate(self.supported_optimizers)}
        assert s["optimizer"] in optimizer_map
        self.loss_combo_box.setCurrentIndex(loss_map[s["loss"]])
        self.optimizer_combo_box.setCurrentIndex(optimizer_map[s["optimizer"]])
        self.try_global_checkbox.setChecked(s["try_global"])
        self.global_max_niter_input.setValue(s["global_max_niter"])
        self.global_niter_success_input.setValue(s["global_niter_success"])
        self.global_step_size_input.setValue(s["global_step_size"])
        self.optimizer_max_niter_input.setValue(s["optimizer_max_niter"])
        self.need_history_checkbox.setChecked(s["need_history"])

    def on_try_global_changed(self):
        checked = self.try_global_checkbox.isChecked()
        self.global_max_niter_input.setEnabled(checked)
        self.global_niter_success_input.setEnabled(checked)
        self.global_step_size_input.setEnabled(checked)

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.setWindowTitle(self.tr("SSU Settings"))
        self.loss_label.setText(self.tr("Loss Function"))
        self.loss_label.setToolTip(self.tr(
            "The function to calculate the difference between prediction and observation."))
        for i, (key, name) in enumerate(self.supported_losses):
            self.loss_combo_box.setItemText(i, name)
        self.optimizer_label.setText(self.tr("Optimizer"))
        self.optimizer_label.setToolTip(self.tr("The optimizer to find the minimum of loss function."))
        for i, (key, name) in enumerate(self.supported_optimizers):
            self.optimizer_combo_box.setItemText(i, name)
        self.try_global_checkbox.setText(self.tr("Global Optimization"))
        self.try_global_checkbox.setToolTip(self.tr("Try global optimization or not."))
        self.global_max_niter_label.setText(self.tr("Maximum Number of Iterations"))
        self.global_max_niter_label.setToolTip(self.tr("Maximum number of iterations of global optimization."))
        self.global_niter_success_label.setText(self.tr("Success Number of Iterations"))
        self.global_niter_success_label.setToolTip(self.tr(
            "The number of iterations that reaching the same minimum value."))
        self.global_step_size_label.setText(self.tr("Step Size"))
        self.global_step_size_label.setToolTip(self.tr("The step size of searching the global minimum."))
        self.optimizer_max_niter_label.setText(self.tr("Maximum Number of Iterations of Optimizer"))
        self.optimizer_max_niter_label.setToolTip(self.tr(
            "Maximum number of iterations of the optimizer in global and local optimization."))
        self.need_history_checkbox.setText(self.tr("Need History"))
        self.need_history_checkbox.setToolTip(self.tr("Record the variation history of parameters or not."))



__all__ = ["SIMPLE_PRESET", "LOESS_PRESET", "LACUSTRINE_PRESET",
           "random_parameters", "random_dataset", "random_sample", "random_mean_sample"]

from typing import *

import numpy as np

from .models import DistributionType, ArtificialDataset

SIMPLE_PRESET = dict(
    target=[
        [(0.0, 0.0), (10.2, 0.0), (1.1, 0.0), (1.0, 0.1)],
        [(0.0, 0.0), (7.5, 0.0), (1.2, 0.0), (2.0, 0.2)],
        [(0.0, 0.0), (5.0, 0.0), (1.0, 0.0), (2.5, 0.5)]],
    distribution_type=DistributionType.SkewNormal)

LOESS_PRESET = dict(
    target=[
        [(0.0, 0.10), (10.2, 0.1), (1.1, 0.1), (1.0, 0.1)],
        [(0.0, 0.10), (7.5, 0.1), (1.2, 0.1), (2.0, 0.1)],
        [(0.0, 0.10), (5.0, 0.2), (1.0, 0.1), (2.5, 0.2)]],
    distribution_type=DistributionType.SkewNormal)

LACUSTRINE_PRESET = dict(
    target=[
        [(0.0, 0.10), (10.2, 0.1), (1.1, 0.1), (1.0, 0.1)],
        [(0.0, 0.10), (7.5, 0.1), (1.2, 0.1), (2.0, 0.1)],
        [(0.0, 0.10), (5.0, 0.2), (1.0, 0.1), (2.5, 0.2)],
        [(0.0, 0.10), (2.5, 0.4), (1.0, 0.1), (1.0, 0.2)]],
    distribution_type=DistributionType.SkewNormal)


def random_parameters(target: Sequence[Sequence[Tuple[float, float]]], n_samples: int):
    """
    Get an array of random parameters which can be used to generate the artificial dataset.

    :param target: It defines the mean and standard deviation of each parameter of each component.
    :param n_samples: The number of samples to generate.
    :return: A numpy array (n_samples, n_parameters, n_components) of the generated parameters.
    """
    n_components = len(target)
    n_parameters = len(target[0])
    parameters = np.random.randn(n_samples, n_parameters, n_components)
    for component_i, sub_target in enumerate(target):
        for param_i, (mean, std) in enumerate(sub_target):
            parameters[:, param_i, component_i] = parameters[:, param_i, component_i] * std + mean
    return parameters


def random_dataset(target: Sequence[Sequence[Tuple[float, float]]], distribution_type: DistributionType, n_samples: int,
                   min_size=0.02, max_size=2000.0, n_classes=101, precision=4, noise=5):
    """
    Generate a random dataset.

    For example, a 3-normal-components `target` is as follows.

    --------------------------- 3 Normal ----------------------------
    |       Location     Scale        Weight                        |
    |       Mean   S.D.  Mean   S.D.  Mean   S.D.                   |
    |    [[(10.2,  0.0), (1.1,  0.0), (1.0,  0.1)],   # Component 1 |
    |     [( 7.5,  0.0), (1.2,  0.0), (2.0,  0.2)],   # Component 2 |
    |     [( 5.0,  0.0), (1.0,  0.0), (2.5,  0.5)]],  # Component 3 |
    -----------------------------------------------------------------

    :param target: It defines the mean and standard deviation of each parameter of each component.
    :param distribution_type: The type of elementary distribution.
    :param n_samples: The number of samples to generate.
    :param min_size: The minimum grain size class in microns.
    :param max_size: The maximum grain size class in microns.
    :param n_classes: The number of grain size classes. More classes can make the distributions more detailed.
    :param precision: The precision level (decimals) of generated grain size distributions.
    :param noise: The standard deviation of the random noise matrix. Usually, should use `precision+1`.
    :return: An instance of the `ArtificialDataset` class.
    """
    parameters = random_parameters(target, n_samples)
    dataset = ArtificialDataset(
        parameters, distribution_type,
        min_size=min_size, max_size=max_size, n_classes=n_classes,
        precision=precision, noise=noise)
    return dataset


def random_sample(target: Sequence[Sequence[Tuple[float, float]]], distribution_type: DistributionType,
                  min_size=0.02, max_size=2000.0, n_classes=101, precision=4, noise=5):
    """
    Generate a random sample.

    :param target: It defines the mean and standard deviation of each parameter of each component.
    :param distribution_type: The type of elementary distribution.
    :param min_size: The minimum grain size class in microns.
    :param max_size: The maximum grain size class in microns.
    :param n_classes: The number of grain size classes. More classes can make the distributions more detailed.
    :param precision: The precision level (decimals) of generated grain size distributions.
    :param noise: The standard deviation of the random noise matrix. Usually, should use `precision+1`.
    :return: An instance of the `ArtificialSample` class.
    """
    parameters = random_parameters(target, 1)
    dataset = ArtificialDataset(
        parameters, distribution_type,
        min_size=min_size, max_size=max_size, n_classes=n_classes,
        precision=precision, noise=noise)
    return dataset[0]


def random_mean_sample(target: Sequence[Sequence[Tuple[float, float]]], distribution_type: DistributionType,
                       min_size=0.02, max_size=2000.0, n_classes=101, precision=4, noise=5):
    """
    Generate a sample using the `mean` values in `target` as the parameters.

    :param target: It defines the mean and standard deviation of each parameter of each component.
    :param distribution_type: The type of elementary distribution.
    :param min_size: The minimum grain size class in microns.
    :param max_size: The maximum grain size class in microns.
    :param n_classes: The number of grain size classes. More classes can make the distributions more detailed.
    :param precision: The precision level (decimals) of generated grain size distributions.
    :param noise: The standard deviation of the random noise matrix. Usually, should use `precision+1`.
    :return: An instance of the `ArtificialSample` class.
    """
    parameters = np.expand_dims(np.array([[mean for (mean, std) in component] for component in target]).T, 0)
    dataset = ArtificialDataset(
        parameters, distribution_type,
        min_size=min_size, max_size=max_size, n_classes=n_classes,
        precision=precision, noise=noise)
    return dataset[0]



__all__ = ["ClusteringAnalyzer"]

import logging
import typing

import matplotlib.pyplot as plt
import numpy as np
from PySide6 import QtCore, QtWidgets, QtGui
from scipy.cluster.hierarchy import fcluster, linkage
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA

from ..statistics import major_statistics
from ..models import Dataset
from ..io import save_clustering
from ..charts.FrequencyGroupChart import FrequencyGroupChart
from ..charts.HierarchicalChart import HierarchicalChart
from ..charts.diagrams import CMDiagramChart


linkages = ("single", "complete", "average", "weighted", "centroid", "median", "ward")
metrics = ("braycurtis", "canberra", "chebyshev", "cityblock", "correlation", "cosine", "dice", "euclidean", "hamming",
           "jaccard", "jensenshannon", "kulsinski", "mahalanobis", "matching", "minkowski", "rogerstanimoto",
           "russellrao", "seuclidean", "sokalmichener", "sokalsneath", "sqeuclidean", "yule")


class ClusteringAnalyzer(QtWidgets.QWidget):
    logger = logging.getLogger("QGrain.ClusteringAnalyzer")

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setWindowTitle(self.tr("Clustering Analyzer"))
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.chart = HierarchicalChart()
        self.main_layout.addWidget(self.chart, 0, 0, 1, 4)
        self.pca_checkbox = QtWidgets.QCheckBox(self.tr("PCA"))
        self.pca_checkbox.setToolTip(self.tr("If use the decomposed features of PCA."))
        self.pca_checkbox.setChecked(True)
        self.pca_ratio_label = QtWidgets.QLabel(self.tr("Variance Ratio"))
        self.pca_ratio_label.setToolTip(self.tr("It determines the numbers of PCs."))
        self.pca_ratio_input = QtWidgets.QDoubleSpinBox()
        self.pca_ratio_input.setDecimals(3)
        self.pca_ratio_input.setRange(0.001, 0.999)
        self.pca_ratio_input.setValue(0.950)
        self.mean_checkbox = QtWidgets.QCheckBox(self.tr("Mean"))
        self.mode_checkbox = QtWidgets.QCheckBox(self.tr("Mode"))
        self.first_percentile_checkbox = QtWidgets.QCheckBox(self.tr("First Percentile (C)"))
        self.median_checkbox = QtWidgets.QCheckBox(self.tr("Median (M)"))
        self.std_checkbox = QtWidgets.QCheckBox(self.tr("Sorting Coefficient"))
        self.skewness_checkbox = QtWidgets.QCheckBox(self.tr("Skewness"))
        self.kurtosis_checkbox = QtWidgets.QCheckBox(self.tr("Kurtosis"))
        self.standardize_checkbox = QtWidgets.QCheckBox(self.tr("Standardize"))
        self.standardize_checkbox.setToolTip(self.tr("Standardization is necessary to balance different features."))
        self.standardize_checkbox.setChecked(True)
        self.main_layout.addWidget(self.pca_checkbox, 1, 0, 1, 2)
        self.main_layout.addWidget(self.pca_ratio_label, 1, 2)
        self.main_layout.addWidget(self.pca_ratio_input, 1, 3)
        self.main_layout.addWidget(self.mean_checkbox, 2, 0)
        self.main_layout.addWidget(self.mode_checkbox, 2, 1)
        self.main_layout.addWidget(self.first_percentile_checkbox, 2, 2)
        self.main_layout.addWidget(self.median_checkbox, 2, 3)
        self.main_layout.addWidget(self.std_checkbox, 3, 0)
        self.main_layout.addWidget(self.skewness_checkbox, 3, 1)
        self.main_layout.addWidget(self.kurtosis_checkbox, 3, 2)
        self.main_layout.addWidget(self.standardize_checkbox, 3, 3)

        self.linkage_label = QtWidgets.QLabel(self.tr("Linkage"))
        self.linkage_label.setToolTip(self.tr("The linkage method for calculating the distance between "
                                              "the newly formed cluster and each observation vector."))
        self.linkage_combo_box = QtWidgets.QComboBox()
        self.linkage_combo_box.addItems(linkages)
        self.linkage_combo_box.setCurrentText("ward")
        self.metric_label = QtWidgets.QLabel(self.tr("Metric"))
        self.metric_label.setToolTip(self.tr("The distance metric."))
        self.metric_combo_box = QtWidgets.QComboBox()
        self.metric_combo_box.addItems(metrics)
        self.metric_combo_box.setCurrentText("euclidean")
        self.main_layout.addWidget(self.linkage_label, 4, 0)
        self.main_layout.addWidget(self.linkage_combo_box, 4, 1)
        self.main_layout.addWidget(self.metric_label, 4, 2)
        self.main_layout.addWidget(self.metric_combo_box, 4, 3)
        self.p_label = QtWidgets.QLabel(self.tr("p"))
        self.p_label.setToolTip(self.tr("The number of leaves at the bottom level of the figure."))
        self.p_input = QtWidgets.QSpinBox()
        self.p_input.setMinimum(1)
        self.main_layout.addWidget(self.p_label, 5, 0)
        self.main_layout.addWidget(self.p_input, 5, 1)
        self.n_clusters_label = QtWidgets.QLabel(self.tr("Number of Clusters"))
        self.n_clusters_label.setToolTip(self.tr("The number of clusters."))
        self.n_clusters_input = QtWidgets.QSpinBox()
        self.n_clusters_input.setMinimum(2)
        self.perform_button = QtWidgets.QPushButton(self.tr("Perform"))
        self.perform_button.clicked.connect(self.perform)
        self.preview_button = QtWidgets.QPushButton(self.tr("Preview"))
        self.preview_button.clicked.connect(self.show_preview)
        self.close_button = QtWidgets.QPushButton(self.tr("Close Preview Charts"))
        self.close_button.clicked.connect(self.close_preview_charts)
        self.show_cm_diagram_button = QtWidgets.QPushButton(self.tr("Show C-M Diagram"))
        self.show_cm_diagram_button.clicked.connect(self.show_cm_diagram)
        self.main_layout.addWidget(self.n_clusters_label, 5, 2)
        self.main_layout.addWidget(self.n_clusters_input, 5, 3)
        self.main_layout.addWidget(self.perform_button, 6, 0)
        self.main_layout.addWidget(self.preview_button, 6, 1)
        self.main_layout.addWidget(self.close_button, 6, 2)
        self.main_layout.addWidget(self.show_cm_diagram_button, 6, 3)
        self.normal_msg = QtWidgets.QMessageBox(self)
        self.file_dialog = QtWidgets.QFileDialog(parent=self)
        self._last_dataset = None
        self._last_result = None
        self._cluster_group_charts: list[FrequencyGroupChart] = []

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    @property
    def linkage_name(self) -> str:
        return self.linkage_combo_box.currentText()

    @property
    def metric_name(self) -> str:
        return self.metric_combo_box.currentText()

    @property
    def p(self) -> int:
        return self.p_input.value()

    @property
    def n_clusters(self) -> int:
        return self.n_clusters_input.value()

    @property
    def data_key(self) -> tuple:
        stat_keys: list[str] = []
        if self.mean_checkbox.isChecked():
            stat_keys.append("mean")
        if self.median_checkbox.isChecked():
            stat_keys.append("median")
        if self.mode_checkbox.isChecked():
            stat_keys.append("mode")
        if self.first_percentile_checkbox.isChecked():
            stat_keys.append("first_percentile")
        if self.std_checkbox.isChecked():
            stat_keys.append("std")
        if self.skewness_checkbox.isChecked():
            stat_keys.append("skewness")
        if self.kurtosis_checkbox.isChecked():
            stat_keys.append("kurtosis")
        if self.pca_checkbox.isChecked() and len(stat_keys) == 0:
            if self.standardize_checkbox.isChecked():
                return f"PCA ({self.pca_ratio_input.value()})", "standardized"
            else:
                return f"PCA ({self.pca_ratio_input.value()})",
        elif (not self.pca_checkbox.isChecked()) and len(stat_keys) > 0:
            if self.standardize_checkbox.isChecked():
                return *stat_keys, "standardized"
            else:
                return stat_keys
        elif len(stat_keys) > 0:
            if self.standardize_checkbox.isChecked():
                return f"PCA ({self.pca_ratio_input.value()})", *stat_keys, "standardized"
            else:
                return f"PCA ({self.pca_ratio_input.value()})", *stat_keys
        else:
            return "GSDs",

    def on_dataset_loaded(self, dataset: Dataset):
        if dataset is None:
            return
        self._last_dataset = dataset
        self._last_result = None
        self.p_input.setMaximum(len(dataset))
        self.n_clusters_input.setMaximum(len(dataset) - 1)
        if len(dataset) > 20:
            self.p_input.setValue(10)
        # self.perform()

    def _get_data_for_clustering(self) -> typing.Tuple[np.ndarray, typing.Tuple]:
        if self._last_dataset is None:
            self.logger.warning("The dataset has not been loaded.")
            return
        data_key = self.data_key
        stat_keys: list[str] = []
        if self.mean_checkbox.isChecked():
            stat_keys.append("mean")
        if self.median_checkbox.isChecked():
            stat_keys.append("median")
        if self.mode_checkbox.isChecked():
            stat_keys.append("mode")
        if self.first_percentile_checkbox.isChecked():
            stat_keys.append("first_percentile")
        if self.std_checkbox.isChecked():
            stat_keys.append("std")
        if self.skewness_checkbox.isChecked():
            stat_keys.append("skewness")
        if self.kurtosis_checkbox.isChecked():
            stat_keys.append("kurtosis")
        if self.pca_checkbox.isChecked() and len(stat_keys) == 0:
            pca = PCA(n_components=self.pca_ratio_input.value())
            transformed = pca.fit_transform(self._last_dataset.distributions)
            if self.standardize_checkbox.isChecked():
                standardized = StandardScaler().fit_transform(transformed)
                return standardized, data_key
            else:
                return transformed, data_key
        elif (not self.pca_checkbox.isChecked()) and len(stat_keys) > 0:
            all_stats = []
            for sample in self._last_dataset:
                stats = major_statistics(sample.classes, sample.classes_phi, sample.distribution,
                                         is_geometric=True, is_fw57=False)
                selected_stats = [stats[key] for key in stat_keys]
                all_stats.append(selected_stats)
            all_stats = np.array(all_stats)
            if self.standardize_checkbox.isChecked():
                standardized = StandardScaler().fit_transform(all_stats)
                return standardized, data_key
            else:
                return all_stats, data_key
        elif self.pca_checkbox.isChecked() and len(stat_keys) > 0:
            pca = PCA(n_components=self.pca_ratio_input.value())
            transformed = pca.fit_transform(self._last_dataset.distributions)
            all_stats = []
            for sample in self._last_dataset:
                stats = major_statistics(sample.classes, sample.classes_phi, sample.distribution,
                                         is_geometric=True, is_fw57=False)
                selected_stats = [stats[key] for key in stat_keys]
                all_stats.append(selected_stats)
            all_stats = np.array(all_stats)
            concatenated = np.concatenate([transformed, all_stats], axis=1)
            if self.standardize_checkbox.isChecked():
                standardized = StandardScaler().fit_transform(concatenated)
                return standardized, data_key
            else:
                return concatenated, data_key
        else:
            return self._last_dataset.distributions, data_key

    def perform(self):
        if self._last_dataset is None:
            self.logger.warning("The dataset has not been loaded.")
            return
        if self._last_result is not None:
            data_key, linkage_name, metric_name, linkage_matrix = self._last_result
            if data_key == self.data_key and linkage_name == self.linkage_name and metric_name == self.metric_name:
                self.logger.debug("The clustering algorithm has been performed on this dataset, "
                                  "use previous result to avoid the repetitive computation.")
                self.chart.show_matrix(linkage_matrix, p=self.p)
                return
        self.logger.debug(
            f"Calculate the linkage matrix with the data key ({self.data_key}), "
            f"method ({self.linkage_name}) and metric ({self.metric_name}).")
        data, data_key = self._get_data_for_clustering()
        try:
            linkage_matrix = linkage(data, method=self.linkage_name, metric=self.metric_name)
            self._last_result = (data_key, self.linkage_name, self.metric_name, linkage_matrix)
            self.chart.show_matrix(linkage_matrix, p=self.p)
        except ValueError as e:
            self.logger.error(f"The linkage method {self.linkage_name} is not compatible with "
                              f"the distance metric {self.metric_name}: {e}.")
            self.show_error(self.tr("The linkage method is not compatible with the distance metric."))
            if self._last_result is not None:
                data_key, linkage_name, metric_name, linkage_matrix = self._last_result
                # avoid unnecessary chart updating
                self.blockSignals(True)
                self.linkage_combo_box.setCurrentText(linkage_name)
                self.metric_combo_box.setCurrentText(metric_name)
                self.blockSignals(False)

    def show_preview(self):
        if self._last_dataset is None:
            self.logger.warning("The dataset has not been loaded.")
            return
        if self._last_result is None:
            self.logger.warning("The clustering has not been performed.")
            return
        self.close_preview_charts()
        data_key, linkage_method, metric, linkage_matrix = self._last_result
        flags = fcluster(linkage_matrix, self.n_clusters, criterion="maxclust")
        self.logger.debug(f"Try to preview the clustering result. Data key: {data_key}. "
                          f"Linkage: {linkage_method}. Metric: {metric}. Number of clusters: {self.n_clusters}.")
        unique_flags = np.unique(flags)
        rect = QtWidgets.QApplication.primaryScreen().availableGeometry()
        x0 = rect.x() + 20
        y0 = rect.y() + 20
        max_w = rect.width()
        max_h = rect.height()
        span = 5
        epoch = 0
        x = x0 + epoch % 10 * 20 + span
        y = y0 + epoch % 10 * 20 + span

        for flag in unique_flags:
            distributions = self._last_dataset.distributions[flags == flag]
            chart = FrequencyGroupChart()
            chart.show_frequency_grop(self._last_dataset.classes_phi, distributions,
                                      title=f"Cluster {flag} (n={len(distributions)})")
            self._cluster_group_charts.append(chart)
            chart.show()
            w = chart.width()
            h = chart.height()
            chart.setGeometry(x, y, w, h)
            QtWidgets.QApplication.processEvents()
            x += w + span
            if x > max_w - w:
                x = x0 + epoch % 10 * 40 + span
                y += h + span
            if y > max_h - h:
                epoch += 1
                x = x0 + epoch % 10 * 40 + span
                y = x0 + epoch % 10 * 40 + span

    def close_preview_charts(self):
        for chart in self._cluster_group_charts:
            chart.setVisible(False)
            chart.close()
        self._cluster_group_charts.clear()

    def show_cm_diagram(self):
        if self._last_dataset is None:
            self.logger.warning("The dataset has not been loaded.")
            return
        if self._last_result is None:
            self.logger.warning("The clustering has not been performed.")
            return
        self.close_preview_charts()
        data_key, linkage_method, metric, linkage_matrix = self._last_result
        flags = fcluster(linkage_matrix, self.n_clusters, criterion="maxclust")
        self.logger.debug(f"Try to show the clusters on a C-M diagram. Data key: {data_key}. "
                          f"Linkage: {linkage_method}. Metric: {metric}. Number of clusters: {self.n_clusters}.")
        unique_flags = np.unique(flags)
        chart = CMDiagramChart()
        marker_size = min(max(2, 5000//len(flags)), 8)
        for i, flag in enumerate(unique_flags):
            key = np.equal(flags, flag)
            samples = [sample for in_cluster, sample in zip(key, self._last_dataset) if in_cluster]
            color = plt.get_cmap("gist_rainbow")(i/len(unique_flags))
            chart.show_samples(samples, append=True, ms=marker_size, mfc=color, mew=0.0)
        chart.show()

    def save_result(self):
        if self._last_dataset is None:
            self.logger.error("The dataset has not been loaded.")
            self.show_error(self.tr("The dataset has not been loaded."))
            return
        if self._last_result is None:
            self.logger.error("The clustering algorithm has not been performed.")
            self.show_error(self.tr("The clustering algorithm has not been performed."))
            return
        filename, _ = self.file_dialog.getSaveFileName(
            self, self.tr("Choose a filename to save the clustering result"), ".", "Microsoft Excel (*.xlsx)")
        if filename is None or filename == "":
            self.logger.info("No filename was selected.")
            return
        progress_dialog = QtWidgets.QProgressDialog(self.tr("Saving the clustering result..."),
                                                    self.tr("Cancel"), 0, 100, self)
        progress_dialog.setWindowTitle("QGrain")
        progress_dialog.setWindowModality(QtCore.Qt.WindowModal)
        try:
            data_key, linkage_method, metric, linkage_matrix = self._last_result
            flags = fcluster(linkage_matrix, self.n_clusters, criterion="maxclust")
            self.logger.debug(f"Try to save the clustering result. Data key: {data_key}. "
                              f"Linkage: {linkage_method}. Metric: {metric}. Number of clusters: {self.n_clusters}.")

            def callback(progress: float):
                if progress_dialog.wasCanceled():
                    raise StopIteration()
                progress_dialog.setValue(int(progress * 100))
                QtCore.QCoreApplication.processEvents()

            save_clustering(self._last_dataset, flags, filename, data_key, linkage_method, metric,
                            progress_callback=callback, logger=self.logger)
        except StopIteration:
            self.logger.info("The saving task was canceled.")
        except Exception as e:
            self.logger.exception(f"An unknown exception was raised: {e}. "
                                  f"Please check the logs for more details.", stack_info=True)
            self.show_error(self.tr("An unknown exception was raised. Please check the logs for more details."))
        finally:
            progress_dialog.close()

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def closeEvent(self, event: QtGui.QCloseEvent):
        self.logger.debug("The main window was closed, close all charts of clusters (if have).")
        for chart in self._cluster_group_charts:
            chart.setVisible(False)
            chart.close()
        self._cluster_group_charts.clear()

    def retranslate(self):
        self.setWindowTitle(self.tr("Clustering Analyzer"))
        self.pca_checkbox.setText(self.tr("PCA"))
        self.pca_checkbox.setToolTip(self.tr("If use the decomposed features of PCA."))
        self.pca_ratio_label.setText(self.tr("Variance Ratio"))
        self.pca_ratio_label.setToolTip(self.tr("It determines the numbers of PCs."))
        self.mean_checkbox.setText(self.tr("Mean"))
        self.mode_checkbox.setText(self.tr("Mode"))
        self.first_percentile_checkbox.setText(self.tr("First Percentile (C)"))
        self.median_checkbox.setText(self.tr("Median (M)"))
        self.std_checkbox.setText(self.tr("Sorting Coefficient"))
        self.skewness_checkbox.setText(self.tr("Skewness"))
        self.kurtosis_checkbox.setText(self.tr("Kurtosis"))
        self.standardize_checkbox.setText(self.tr("Standardize"))
        self.standardize_checkbox.setToolTip(self.tr("Standardization is necessary to balance different features."))
        self.linkage_label.setText(self.tr("Linkage"))
        self.linkage_label.setToolTip(self.tr("The linkage method for calculating the distance between "
                                              "the newly formed cluster and each observation vector."))
        self.metric_label.setText(self.tr("Metric"))
        self.metric_label.setToolTip(self.tr("The distance metric."))
        self.p_label.setText(self.tr("p"))
        self.p_label.setToolTip(self.tr("The number of leaves at the bottom level of the figure."))
        self.n_clusters_label.setText(self.tr("Number of Clusters"))
        self.n_clusters_label.setToolTip(self.tr("The number of clusters."))
        self.perform_button.setText(self.tr("Perform"))
        self.preview_button.setText(self.tr("Preview"))
        self.close_button.setText(self.tr("Close Preview Charts"))
        self.show_cm_diagram_button.setText(self.tr("Show C-M Diagram"))



__all__ = [
    "DiagramChart",
    "Folk54GSMDiagramChart",
    "Folk54SSCDiagramChart",
    "BP12GSMDiagramChart",
    "BP12SSCDiagramChart",
    "CMDiagramChart"]

from typing import *

import numpy as np
from matplotlib.path import Path
from matplotlib.patches import Circle, PathPatch

from . import BaseChart
from . import normal_color, background_color
from ..models import Sample
from ..statistics import cm, proportions_gsm, proportions_ssc, to_phi, to_microns


class DiagramChart(BaseChart):
    def __init__(self, parent=None, figsize=(4.4, 3.1)):
        super().__init__(parent=parent, figsize=figsize)
        self.axes = self._figure.subplots()
        self.draw_base()
        self._sample_batches = []

    def trans_pos(self, a, b) -> Tuple[float, float]:
        pass

    @property
    def title(self):
        return ""

    @property
    def lines(self):
        return []

    @property
    def labels(self):
        return []

    @property
    def n_batches(self) -> int:
        return len(self._sample_batches)

    def plot_legend(self):
        pass

    def draw_base(self):
        self.axes.axis("off")
        self.axes.set_aspect(1)
        # self.axes.set_title(self.title)

        for a, b, kwargs in self.lines:
            x, y = self.trans_pos(a, b)
            self.axes.plot(x, y, **kwargs, label="_")

        for (a, b), text, kwargs in self.labels:
            x, y = self.trans_pos(a, b)
            self.axes.text(x, y, text, color=normal_color(), label="_", **kwargs)

        self.plot_legend()
        self._figure.tight_layout()

    def convert_samples(self, samples: List[Sample]) -> Tuple[Sequence[float], Sequence[float]]:
        pass

    def show_samples(self, samples: List[Sample],
                     append=False, c="#ffffff00",
                     marker=".", ms=8, mfc="red", mew=0.0,
                     **kwargs):
        if len(samples) == 0:
            return
        if not append:
            self.axes.clear()
            self.draw_base()
            self._sample_batches.clear()
        a, b = self.convert_samples(samples)
        x, y = self.trans_pos(a, b)
        self.axes.plot(x, y, c=c, marker=marker, ms=ms, mfc=mfc, mew=mew, label=f"batch_{self.n_batches}",
                       zorder=100, **kwargs)
        self._canvas.draw()
        plot_kwargs = dict(c=c, marker=marker, ms=ms, mfc=mfc, mew=mew)
        plot_kwargs.update(kwargs)
        self._sample_batches.append((samples, plot_kwargs))

    def update_chart(self):
        self._figure.clear()
        self.axes = self._figure.subplots()
        self.draw_base()
        for i, (samples, plot_kwargs) in enumerate(self._sample_batches):
            a, b = self.convert_samples(samples)
            x, y = self.trans_pos(a, b)
            self.axes.plot(x, y, label=f"batch_{i}", **plot_kwargs)
        self._canvas.draw()


class Folk54GSMDiagramChart(DiagramChart):
    def __init__(self, parent=None, figsize=(6, 4.5)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("GSM Diagram (Folk, 1954)"))

    @property
    def title(self):
        return "Gravel-sand-mud diagram (Folk, 1954)"

    @property
    def lines(self):
        STRUCTURAL_LINES = [
            ([0.0, 0.0, 1.0, 0.0], [0.0, 1.0, 0.0, 0.0], dict(c=normal_color(), linewidth=0.8)),
            # the 3 sides of this equilateral triangle
            ([0.0, 1.0], [0.01, 0.01], dict(c=normal_color(), linewidth=0.8)),  # gravel = Trace
            ([0.0, 1.0], [0.05, 0.05], dict(c=normal_color(), linewidth=0.8)),  # gravel = 5%
            ([0.0, 1.0], [0.3, 0.3], dict(c=normal_color(), linewidth=0.8)),  # gravel = 30%
            ([0.0, 1.0], [0.8, 0.8], dict(c=normal_color(), linewidth=0.8)),  # gravel = 80%

            ([1 / 10, 1 / 10], [0.0, 0.05], dict(c=normal_color(), linewidth=0.8)),  # sand: mud = 1:9
            ([1 / 2, 1 / 2], [0.0, 0.8], dict(c=normal_color(), linewidth=0.8)),  # sand: mud = 1:1
            ([9 / 10, 9 / 10], [0.0, 0.8], dict(c=normal_color(), linewidth=0.8))]  # sand: mud = 9:1
        ADDITIONAL_LINES = [
            ([0.05, -0.03], [0.02, 0.12], dict(c=normal_color(), linewidth=0.8)),
            ([0.3, 0.3], [0.02, 0.06], dict(c=normal_color(), linewidth=0.8)),
            ([0.7, 0.7], [0.02, 0.06], dict(c=normal_color(), linewidth=0.8)),
            ([0.95, 1.03], [0.02, 0.12], dict(c=normal_color(), linewidth=0.8)),
            ([0.95, 1.03], [0.02, 0.12], dict(c=normal_color(), linewidth=0.8)),
            ([0.95, 1.03], [0.2, 0.275], dict(c=normal_color(), linewidth=0.8)),
            ([0.95, 1.03], [0.5, 0.55], dict(c=normal_color(), linewidth=0.8))]

        return STRUCTURAL_LINES + ADDITIONAL_LINES

    @property
    def labels(self):
        s = 0.01
        LABELS = [((-s, -s), "Mud", dict(ha="right", va="top", fontsize=8, fontweight="bold")),
                  ((0.5, 1 + s), "Gravel", dict(ha="center", va="bottom", fontsize=8, fontweight="bold")),
                  ((1 + s, -s), "Sand", dict(ha="left", va="top", fontsize=8, fontweight="bold")),
                  ((-5 * s, 0.55), "Gravel %", dict(ha="right", va="center", fontsize=8, fontweight="bold")),
                  ((0.5, -5 * s), "Sand:Mud Ratio", dict(ha="center", va="top", fontsize=8, fontweight="bold")),

                  ((0.05, -s), "Mud", dict(ha="center", va="top", fontsize=7)),
                  ((0.3, -s), "Sandy Mud", dict(ha="center", va="top", fontsize=7)),
                  ((0.7, -s), "Muddy Sand", dict(ha="center", va="top", fontsize=7)),
                  ((0.95, -s), "Sand", dict(ha="center", va="top", fontsize=7)),

                  ((-0.03, 0.12), "Slightly\nGravelly\nMud", dict(ha="right", va="bottom", fontsize=7)),
                  ((0.3, 0.08), "Slightly Gravelly Sandy Mud", dict(ha="center", va="bottom", fontsize=7)),
                  ((0.7, 0.08), "Slightly Gravelly Muddy Sand", dict(ha="center", va="bottom", fontsize=7)),
                  ((1.03, 0.12), "Slightly\nGravelly\nSand", dict(ha="left", va="bottom", fontsize=7)),

                  ((0.2, 0.2), "Gravelly Mud", dict(ha="center", va="center", fontsize=7)),
                  ((0.7, 0.2), "Gravelly Muddy Sand", dict(ha="center", va="center", fontsize=7)),
                  ((1.03, 0.28), "Gravelly\nSand", dict(ha="left", va="bottom", fontsize=7)),
                  ((0.2, 0.55), "Muddy\nGravel", dict(ha="center", va="center", fontsize=7)),
                  ((0.7, 0.55), "Muddy\nSandy\nGravel", dict(ha="center", va="center", fontsize=7)),
                  ((1.03, 0.55), "Sandy\nGravel", dict(ha="left", va="bottom", fontsize=7)),
                  ((0.5, 0.85), "Gravel", dict(ha="center", va="bottom", fontsize=7)),

                  ((-1 * s, 0.01), "Trace", dict(ha="right", va="center", fontsize=7)),
                  ((-1 * s, 0.05), "5%", dict(ha="right", va="center", fontsize=7)),
                  ((-3 * s, 0.3), "30%", dict(ha="right", va="center", fontsize=7)),
                  ((-8 * s, 0.8), "80%", dict(ha="right", va="center", fontsize=7)),

                  ((1 / 10, -s), "1:9", dict(ha="center", va="top", fontsize=7)),
                  ((1 / 2, -s), "1:1", dict(ha="center", va="top", fontsize=7)),
                  ((9 / 10, -s), "9:1", dict(ha="center", va="top", fontsize=7))]

        return LABELS

    def trans_pos(self, sand, gravel):
        sand, gravel = np.array(sand), np.array(gravel)
        y = 0.5 * np.sqrt(3) * gravel
        # calculate the cross of two lines
        # first line: y = 0.5 * np.sqrt(3) * gravel
        # second line: the line cross the two points, (sand, 0) and (0.5, 0.5*np.sqrt(3))
        key = np.not_equal(sand, 0.5)
        x = np.full_like(y, np.nan)
        a = np.full_like(y, np.nan)
        b = np.full_like(y, np.nan)
        x[~key] = 0.5
        a[key] = 0.5 * np.sqrt(3) / (0.5 - sand[key])
        b[key] = -a[key] * sand[key]
        x[key] = (y[key] - b[key]) / a[key]
        assert not np.any(np.isnan(x))
        return x, y

    def convert_samples(self, samples: List[Sample]) -> Tuple[Sequence[float], Sequence[float]]:
        sand = []
        gravel = []
        for i, sample in enumerate(samples):
            gravel_i, sand_i, mud_i = proportions_gsm(sample.classes_phi, sample.distribution)
            sand.append(sand_i)
            gravel.append(gravel_i)
        return sand, gravel

    def retranslate(self):
        self.setWindowTitle(self.tr("GSM Diagram (Folk, 1954)"))
        self.edit_figure_action.setText(self.tr("Edit Figure"))
        self.save_figure_action.setText(self.tr("Save Figure"))


class Folk54SSCDiagramChart(DiagramChart):
    def __init__(self, parent=None, figsize=(6, 4.5)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("SSC Diagram (Folk, 1954)"))

    @property
    def title(self):
        return "Sand-silt-clay diagram (Folk, 1954)"

    @property
    def lines(self):
        STRUCTURAL_LINES = [
            ([0.0, 0.0, 1.0, 0.0], [0.0, 1.0, 0.0, 0.0], dict(c=normal_color(), linewidth=0.8)),
            # the 3 sides of this equilateral triangle
            ([0.0, 1.0], [0.1, 0.1], dict(c=normal_color(), linewidth=0.8)),  # sand = 10%
            ([0.0, 1.0], [0.5, 0.5], dict(c=normal_color(), linewidth=0.8)),  # sand = 50%
            ([0.0, 1.0], [0.9, 0.9], dict(c=normal_color(), linewidth=0.8)),  # sand = 90%
            ([1 / 3, 1 / 3], [0.0, 0.9], dict(c=normal_color(), linewidth=0.8)),  # clay: silt = 1:2
            ([2 / 3, 2 / 3], [0.0, 0.9], dict(c=normal_color(), linewidth=0.8))]  # clay: silt = 2:1
        ADDITIONAL_LINES = []

        return STRUCTURAL_LINES + ADDITIONAL_LINES

    @property
    def labels(self):
        s = 0.01
        LABELS = [
            ((-s, -s), "Clay", dict(ha="right", va="top", fontsize=8, fontweight="bold")),
            ((0.5, 1 + s), "Sand", dict(ha="center", va="bottom", fontsize=8, fontweight="bold")),
            ((1 + s, -s), "Silt", dict(ha="left", va="top", fontsize=8, fontweight="bold")),
            ((-5 * s, 0.55), "Sand %", dict(ha="right", va="center", fontsize=8, fontweight="bold")),
            ((0.5, -3 * s), "Silt:Clay Ratio", dict(ha="center", va="top", fontsize=8, fontweight="bold")),

            ((0.16, 0.05), "Clay", dict(ha="center", va="center", fontsize=7)),
            ((0.5, 0.05), "Mud", dict(ha="center", va="center", fontsize=7)),
            ((0.84, 0.05), "Silt", dict(ha="center", va="center", fontsize=7)),
            ((0.16, 0.3), "Sandy Clay", dict(ha="center", va="center", fontsize=7)),
            ((0.5, 0.3), "Sandy Mud", dict(ha="center", va="center", fontsize=7)),
            ((0.84, 0.3), "Sandy Silt", dict(ha="center", va="center", fontsize=7)),
            ((0.16, 0.55), "Clayey Sand", dict(ha="center", va="center", fontsize=7)),
            ((0.5, 0.55), "Muddy Sand", dict(ha="center", va="center", fontsize=7)),
            ((0.84, 0.55), "Silty Sand", dict(ha="center", va="center", fontsize=7)),
            ((0.5, 0.91), "Sand", dict(ha="center", va="bottom", fontsize=7)),

            ((-1 * s, 0.1), "10%", dict(ha="right", va="center", fontsize=7)),
            ((-5 * s, 0.5), "50%", dict(ha="right", va="center", fontsize=7)),
            ((-9 * s, 0.9), "90%", dict(ha="right", va="center", fontsize=7)),

            ((1 / 3, -s), "1:2", dict(ha="center", va="top", fontsize=7)),
            ((2 / 3, -s), "2:1", dict(ha="center", va="top", fontsize=7))]

        return LABELS

    def trans_pos(self, silt, sand):
        silt, sand = np.array(silt), np.array(sand)
        y = 0.5 * np.sqrt(3) * sand
        # calculate the cross of two lines
        # first line: 0.5 * np.sqrt(3) * sand
        # second line: the line cross the two points, (clay, 0) and (0.5, 0.5*np.sqrt(3))
        # y = ax + b
        key = np.not_equal(silt, 0.5)
        x = np.full_like(y, np.nan)
        a = np.full_like(y, np.nan)
        b = np.full_like(y, np.nan)
        x[~key] = 0.5
        a[key] = 0.5 * np.sqrt(3) / (0.5 - silt[key])
        b[key] = -a[key] * silt[key]
        x[key] = (y[key] - b[key]) / a[key]
        assert not np.any(np.isnan(x))
        return x, y

    def convert_samples(self, samples: List[Sample]) -> Tuple[Sequence[float], Sequence[float]]:
        silt = []
        sand = []
        for i, sample in enumerate(samples):
            sand_i, silt_i, clay_i = proportions_ssc(sample.classes_phi, sample.distribution)
            silt.append(silt_i)
            sand.append(sand_i)
        return silt, sand

    def retranslate(self):
        self.setWindowTitle(self.tr("SSC Diagram (Folk, 1954)"))
        self.edit_figure_action.setText(self.tr("Edit Figure"))
        self.save_figure_action.setText(self.tr("Save Figure"))


class BP12GSMDiagramChart(DiagramChart):
    def __init__(self, parent=None, figsize=(6, 4.5)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("GSM Diagram (Blott & Pye, 2012)"))

    @property
    def title(self):
        return "Gravel-sand-mud diagram (Blott & Pye, 2012)"

    @property
    def lines(self):
        span = 0.01
        STRUCTURAL_LINES = [
            ([0.0, 0.0, 1.0, 0.0], [0.0, 1.0, 0.0, 0.0], dict(c=normal_color(), linewidth=0.8)),
            # the 3 sides of this equilateral triangle
            # Gravel
            ([0.0, 0.99], [0.01, 0.01], dict(c=normal_color(), linewidth=0.8)),  # gravel = Trace
            ([0.0, 0.95], [0.05, 0.05], dict(c=normal_color(), linewidth=0.8)),  # gravel = 5%
            ([0.0, 0.8], [0.2, 0.2], dict(c=normal_color(), linewidth=0.8)),  # gravel = 20%
            ([0.0, 1 / 3, 0.5], [0.5, 1 / 3, 0.5], dict(c=normal_color(), linewidth=0.8)),  # gravel = 50%, 33%
            # Sand
            ([0.01, 0.01], [0.0, 0.99], dict(c=normal_color(), linewidth=0.8)),  # sand = Trace
            ([0.05, 0.05], [0.0, 0.95], dict(c=normal_color(), linewidth=0.8)),  # sand = 5%
            ([0.2, 0.2], [0.0, 0.8], dict(c=normal_color(), linewidth=0.8)),  # sand = 20%
            ([0.5, 1 / 3, 0.5], [0.0, 1 / 3, 0.5], dict(c=normal_color(), linewidth=0.8)),  # sand = 50%, 33%
            # Mud
            ([0.99, 0.0], [0.0, 0.99], dict(c=normal_color(), linewidth=0.8)),  # mud = Trace
            ([0.95, 0.0], [0.0, 0.95], dict(c=normal_color(), linewidth=0.8)),  # mud = 5%
            ([0.8, 0.0], [0.0, 0.8], dict(c=normal_color(), linewidth=0.8)),  # mud = 20%
            ([0.5, 1 / 3, 0.0], [0.0, 1 / 3, 0.5], dict(c=normal_color(), linewidth=0.8))]  # mud = 50%, 33%
        ADDITIONAL_LINES = [
            ([0.0, -span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),  # M
            ([0.0, -2 * span], [1.0, 1 + 4 * span], dict(c=normal_color(), linewidth=0.4)),  # G
            ([1.0, 1 + 2 * span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),  # S
            # ticks of Gravel
            ([0.0, -span], [1.0, 1.0 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.9, 0.9 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.8, 0.8 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.7, 0.7 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.6, 0.6 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.5, 0.5 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.4, 0.4 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.3, 0.3 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.2, 0.2 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.1, 0.1 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -span], [0.0, span], dict(c=normal_color(), linewidth=0.4)),
            # ticks of Sand
            ([0.0, 0.0 + 0.1 * span], [1.0, 1.0 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.1, 0.1 + 0.1 * span], [0.9, 0.9 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.2, 0.2 + 0.1 * span], [0.8, 0.8 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.3, 0.3 + 0.1 * span], [0.7, 0.7 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.4, 0.4 + 0.1 * span], [0.6, 0.6 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.5, 0.5 + 0.1 * span], [0.5, 0.5 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.6, 0.6 + 0.1 * span], [0.4, 0.4 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.7, 0.7 + 0.1 * span], [0.3, 0.3 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.8, 0.8 + 0.1 * span], [0.2, 0.2 + span], dict(c=normal_color(), linewidth=0.4)),
            ([0.9, 0.9 + 0.1 * span], [0.1, 0.1 + span], dict(c=normal_color(), linewidth=0.4)),
            ([1.0, 1.0 + 0.1 * span], [0.0, span], dict(c=normal_color(), linewidth=0.4)),
            # ticks of Mud
            ([0.0, span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.1, 0.1 + span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.2, 0.2 + span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.3, 0.3 + span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.4, 0.4 + span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.5, 0.5 + span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.6, 0.6 + span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.7, 0.7 + span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.8, 0.8 + span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([0.9, 0.9 + span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),
            ([1.0, 1.0 - span], [0.0, -span], dict(c=normal_color(), linewidth=0.4)),

            # guide lines
            ([0.03, 0.03 + 4 * span], [0.0, -2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.13, 0.13 + 4 * span], [0.0, -2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.33, 0.33 + 4 * span], [0.0, -2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.63, 0.63 + 4 * span], [0.0, -2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.83, 0.83 + 4 * span], [0.0, -2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.97, 0.97], [0.0, -2 * span], dict(c=normal_color(), linewidth=0.4)),

            ([0.0, -2 * span], [0.03, 0.03 + 2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2 * span], [0.13, 0.13 + 2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2 * span], [0.33, 0.33 + 2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2 * span], [0.63, 0.63 + 2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2 * span], [0.83, 0.83 + 2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2 * span], [0.97, 0.97], dict(c=normal_color(), linewidth=0.4)),

            ([0.97, 0.97 + 0.2 * span], [0.03, 0.03 + 2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.87, 0.87 + 0.2 * span], [0.13, 0.13 + 2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.67, 0.67 + 0.2 * span], [0.33, 0.33 + 2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.37, 0.37 + 0.2 * span], [0.63, 0.63 + 2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.17, 0.17 + 0.2 * span], [0.83, 0.83 + 2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.03, 0.03 + 0.2 * span], [0.97, 0.97 + 2 * span], dict(c=normal_color(), linewidth=0.4)),

            ([0.03, 0.03 + 2 * span], [0.95, 0.95 + 2 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.03, -2 * span], [0.03, 0.03 + 4 * span], dict(c=normal_color(), linewidth=0.4)),
            ([0.94, 0.94 + 0.4 * span], [0.03, 0.03 + 4 * span], dict(c=normal_color(), linewidth=0.4)),

        ]
        return STRUCTURAL_LINES + ADDITIONAL_LINES

    @property
    def labels(self):
        s = 0.01
        LABELS = [
            ((-s, -s), "M", dict(ha="right", va="top", fontsize=8, fontweight="bold")),
            ((-2 * s, 1 + 4 * s), "G", dict(ha="center", va="bottom", fontsize=8, fontweight="bold")),
            ((1 + 2 * s, -s), "S", dict(ha="left", va="top", fontsize=8, fontweight="bold")),
            ((-5 * s, 0.55), "Gravel %", dict(ha="right", va="center", fontsize=8, fontweight="bold")),
            ((0.5, 0.55), "Sand %", dict(ha="left", va="center", fontsize=8, fontweight="bold")),
            ((0.5 + 5 * s, -5 * np.sqrt(3) * s), "Mud %", dict(ha="center", va="top", fontsize=8, fontweight="bold")),
            # tick labels of Gravel
            ((-s, 1.0 + s), "100%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.9 + s), "90%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.8 + s), "80%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.7 + s), "70%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.6 + s), "60%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.5 + s), "50%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.4 + s), "40%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.3 + s), "30%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.2 + s), "20%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.1 + s), "10%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.0 + s), "0%", dict(ha="right", va="bottom", fontsize=7)),
            # tick labels of Sand
            ((0.0, 1.0 + s), "0%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.1, 0.9 + s), "10%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.2, 0.8 + s), "20%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.3, 0.7 + s), "30%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.4, 0.6 + s), "40%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.5, 0.5 + s), "50%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.6, 0.4 + s), "60%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.7, 0.3 + s), "70%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.8, 0.2 + s), "80%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.9, 0.1 + s), "90%", dict(ha="left", va="bottom", fontsize=7)),
            ((1.0, 0.0 + s), "100%", dict(ha="left", va="bottom", fontsize=7)),
            # tick labels of Mud
            ((2 * s, -s), "100%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.1 + 2 * s, -s), "90%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.2 + 2 * s, -s), "80%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.3 + 2 * s, -s), "70%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.4 + 2 * s, -s), "60%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.5 + 2 * s, -s), "50%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.6 + 2 * s, -s), "40%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.7 + 2 * s, -s), "30%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.8 + 2 * s, -s), "20%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.9 + 2 * s, -s), "10%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((1.0 - s, -s), "0%", dict(ha="center", va="top", fontsize=7, rotation=-45)),

            ((0.125, 0.025), "(vg)(s)M", dict(ha="center", va="center", fontsize=6)),
            ((0.35, 0.025), "(vg)sM", dict(ha="center", va="center", fontsize=6)),
            ((0.625, 0.025), "(vg)mS", dict(ha="center", va="center", fontsize=6)),
            ((0.850, 0.025), "(vg)(m)S", dict(ha="center", va="center", fontsize=6)),
            ((0.125, 0.125), "(g)(s)M", dict(ha="center", va="center", fontsize=6)),
            ((0.3, 0.125), "(g)sM", dict(ha="center", va="center", fontsize=6)),
            ((0.575, 0.125), "(g)mS", dict(ha="center", va="center", fontsize=6)),
            ((0.750, 0.125), "(g)(m)S", dict(ha="center", va="center", fontsize=6)),
            ((0.125, 0.35), "(s)gM", dict(ha="center", va="center", fontsize=6)),
            ((0.3, 0.3), "gsM", dict(ha="center", va="center", fontsize=6)),
            ((0.4, 0.3), "gmS", dict(ha="center", va="center", fontsize=6)),
            ((0.525, 0.35), "(m)gS", dict(ha="center", va="center", fontsize=6)),
            ((0.125, 0.55), "(s)mG", dict(ha="center", va="center", fontsize=6)),
            ((1 / 3 - (0.45 - 1 / 3) / 2, 0.45), "smG", dict(ha="center", va="center", fontsize=6)),
            ((0.325, 0.55), "(m)sG", dict(ha="center", va="center", fontsize=6)),
            ((0.125, 0.75), "(s)(m)G", dict(ha="center", va="center", fontsize=6)),

            ((0.03, 0.125), "(vs)(g)M", dict(ha="center", va="center", fontsize=6, rotation=60)),
            ((0.03, 0.35), "(vs)gM", dict(ha="center", va="center", fontsize=6, rotation=60)),
            ((0.03, 0.6), "(vs)mG", dict(ha="center", va="center", fontsize=6, rotation=60)),
            ((0.03, 0.85), "(vs)(m)G", dict(ha="center", va="center", fontsize=6, rotation=60)),

            ((0.845, 0.125), "(vm)(g)S", dict(ha="center", va="center", fontsize=6, rotation=-60)),
            ((0.62, 0.35), "(vm)gS", dict(ha="center", va="center", fontsize=6, rotation=-60)),
            ((0.37, 0.6), "(vm)sG", dict(ha="center", va="center", fontsize=6, rotation=-60)),
            ((0.12, 0.85), "(vm)(s)G", dict(ha="center", va="center", fontsize=6, rotation=-60)),

            ((0.05 + 4 * s, -2 * s), "(vs)M", dict(ha="center", va="top", fontsize=6, rotation=-45)),
            ((0.14 + 4 * s, -2 * s), "(s)M", dict(ha="center", va="top", fontsize=6)),
            ((0.34 + 4 * s, -2 * s), "sM", dict(ha="center", va="top", fontsize=6)),
            ((0.64 + 4 * s, -2 * s), "mS", dict(ha="center", va="top", fontsize=6)),
            ((0.84 + 4 * s, -2 * s), "(m)S", dict(ha="center", va="top", fontsize=6)),
            ((0.94 + 4 * s, -2 * s), "(vm)S", dict(ha="center", va="top", fontsize=6, rotation=-45)),

            ((-2 * s, 0.03 + 2 * s), "(vg)M", dict(ha="right", va="center", fontsize=6)),
            ((-2 * s, 0.13 + 2 * s), "(g)M", dict(ha="right", va="bottom", fontsize=6)),
            ((-2 * s, 0.33 + 2 * s), "gM", dict(ha="right", va="bottom", fontsize=6)),
            ((-2 * s, 0.63 + 2 * s), "mG", dict(ha="right", va="bottom", fontsize=6)),
            ((-2 * s, 0.83 + 2 * s), "(m)G", dict(ha="right", va="bottom", fontsize=6)),
            ((-2 * s, 0.97), "(vm)G", dict(ha="right", va="center", fontsize=6)),

            ((0.97 + 0.2 * s, 0.03 + 2 * s), "(vg)S", dict(ha="left", va="center", fontsize=6)),
            ((0.87 + 0.2 * s, 0.13 + 2 * s), "(g)S", dict(ha="left", va="bottom", fontsize=6)),
            ((0.67 + 0.2 * s, 0.33 + 2 * s), "gS", dict(ha="left", va="bottom", fontsize=6)),
            ((0.37 + 0.2 * s, 0.63 + 2 * s), "sG", dict(ha="left", va="bottom", fontsize=6)),
            ((0.17 + 0.2 * s, 0.83 + 2 * s), "(s)G", dict(ha="left", va="bottom", fontsize=6)),
            ((0.03 + 0.2 * s, 0.97 + 2 * s), "(vs)G", dict(ha="left", va="bottom", fontsize=6)),

            ((0.03 + 0.5 * s, 0.97 + 2 * s), "(vs)(vm)G", dict(ha="left", va="top", fontsize=6)),
            ((-2 * s, 0.03 + 4 * s), "(vg)(vs)G", dict(ha="right", va="bottom", fontsize=6)),
            ((0.94 + 0.4 * s, 0.03 + 4 * s), "(vg)(vm)G", dict(ha="left", va="bottom", fontsize=6))]

        return LABELS

    def plot_legend(self):
        x0, y0 = 0.7, 0.9
        w, h = 0.5, 0.16
        span = 0.01
        row_h = 0.02
        self.axes.plot([x0, x0 + w, x0 + w, x0, x0], [y0, y0, y0 - h, y0 - h, y0], c=normal_color(), linewidth=0.8)
        self.axes.text(x0 + w / 2, y0 - span, "Conventions", ha="center", va="top", fontsize=6,
                       color=normal_color())
        texts = ["UPPER CASE - Largest component (noun)",
                 "Lower case - Descriptive term (adjective)",
                 "( ) - Slightly (qualification)",
                 "(v ) - Very slightly (qualification)"]
        for row, text in enumerate(texts, 1):
            x = x0 + span
            y = y0 - span - row * (row_h + span)
            self.axes.text(x, y, text, ha="left", va="top", fontsize=6, color=normal_color())

        x0, y0 = 0.9, 0.7
        w, h = 0.3, 0.32
        span = 0.01
        row_h = 0.02
        self.axes.plot([x0, x0 + w, x0 + w, x0, x0], [y0, y0, y0 - h, y0 - h, y0], c=normal_color(), linewidth=0.8)
        self.axes.text(x0 + w / 2, y0 - span, "Term used", ha="center", va="top", fontsize=6,
                       color=normal_color())
        texts = ["G - Gravel    g - Gravelly",
                 "S - Sand      s - Sandy",
                 "M - Mud       m - Muddy",
                 "(g)  - Slightly gravelly",
                 "(s)  - Slightly sandy",
                 "(m)  - Slightly muddy",
                 "(vg) - Very slightly gravelly",
                 "(vs) - Very slightly sandy",
                 "(vm) - Very slightly muddy"]
        for row, text in enumerate(texts, 1):
            x = x0 + span
            y = y0 - span - row * (row_h + span)
            self.axes.text(x, y, text, ha="left", va="top", fontsize=6, color=normal_color())

    def trans_pos(self, a, b):
        a, b = np.array(a), np.array(b)
        y = 0.5 * np.sqrt(3) * b
        x = 0.5 * b + a
        return x, y

    def convert_samples(self, samples: List[Sample]) -> Tuple[Sequence[float], Sequence[float]]:
        sand = []
        gravel = []
        for i, sample in enumerate(samples):
            gravel_i, sand_i, mud_i = proportions_gsm(sample.classes_phi, sample.distribution)
            sand.append(sand_i)
            gravel.append(gravel_i)
        return sand, gravel

    def retranslate(self):
        self.setWindowTitle(self.tr("GSM Diagram (Blott & Pye, 2012)"))
        self.edit_figure_action.setText(self.tr("Edit Figure"))
        self.save_figure_action.setText(self.tr("Save Figure"))


class BP12SSCDiagramChart(DiagramChart):
    def __init__(self, parent=None, figsize=(6, 4.5)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("SSC Diagram (Blott & Pye, 2012)"))

    @property
    def title(self):
        return "Sand-silt-clay diagram (Blott & Pye, 2012)"

    @property
    def lines(self):
        s = 0.01
        STRUCTURAL_LINES = [
            ([0.0, 0.0, 1.0, 0.0], [0.0, 1.0, 0.0, 0.0], dict(c=normal_color(), linewidth=0.8)),
            # the 3 sides of this equilateral triangle
            # Sand
            ([0.0, 0.99], [0.01, 0.01], dict(c=normal_color(), linewidth=0.8)),  # sand = Trace
            ([0.0, 0.95], [0.05, 0.05], dict(c=normal_color(), linewidth=0.8)),  # sand = 5%
            ([0.0, 0.8], [0.2, 0.2], dict(c=normal_color(), linewidth=0.8)),  # sand = 20%
            ([0.0, 1 / 3, 0.5], [0.5, 1 / 3, 0.5], dict(c=normal_color(), linewidth=0.8)),  # sand = 50%, 33%
            # Silt
            ([0.01, 0.01], [0.0, 0.99], dict(c=normal_color(), linewidth=0.8)),  # silt = Trace
            ([0.05, 0.05], [0.0, 0.95], dict(c=normal_color(), linewidth=0.8)),  # silt = 5%
            ([0.2, 0.2], [0.0, 0.8], dict(c=normal_color(), linewidth=0.8)),  # silt = 20%
            ([0.5, 1 / 3, 0.5], [0.0, 1 / 3, 0.5], dict(c=normal_color(), linewidth=0.8)),  # silt = 50%, 33%
            # Clay
            ([0.99, 0.0], [0.0, 0.99], dict(c=normal_color(), linewidth=0.8)),  # clay = Trace
            ([0.95, 0.0], [0.0, 0.95], dict(c=normal_color(), linewidth=0.8)),  # clay = 5%
            ([0.8, 0.0], [0.0, 0.8], dict(c=normal_color(), linewidth=0.8)),  # clay = 20%
            ([0.5, 1 / 3, 0.0], [0.0, 1 / 3, 0.5], dict(c=normal_color(), linewidth=0.8))]  # clay = 50%, 33%
        ADDITIONAL_LINES = [
            ([0.0, -s], [0.0, -s], dict(c=normal_color(), linewidth=0.4)),  # S
            ([0.0, -2 * s], [1.0, 1 + 4 * s], dict(c=normal_color(), linewidth=0.4)),  # SI
            ([1.0, 1 + 2 * s], [0.0, -s], dict(c=normal_color(), linewidth=0.4)),  # C
            # ticks of Sand
            ([0.0, -s], [1.0, 1.0 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -s], [0.9, 0.9 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -s], [0.8, 0.8 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -s], [0.7, 0.7 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -s], [0.6, 0.6 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -s], [0.5, 0.5 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -s], [0.4, 0.4 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -s], [0.3, 0.3 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -s], [0.2, 0.2 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -s], [0.1, 0.1 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -s], [0.0, s], dict(c=normal_color(), linewidth=0.4)),
            # ticks of Silt
            ([0.0, 0.0 + 0.1 * s], [1.0, 1.0 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.1, 0.1 + 0.1 * s], [0.9, 0.9 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.2, 0.2 + 0.1 * s], [0.8, 0.8 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.3, 0.3 + 0.1 * s], [0.7, 0.7 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.4, 0.4 + 0.1 * s], [0.6, 0.6 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.5, 0.5 + 0.1 * s], [0.5, 0.5 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.6, 0.6 + 0.1 * s], [0.4, 0.4 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.7, 0.7 + 0.1 * s], [0.3, 0.3 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.8, 0.8 + 0.1 * s], [0.2, 0.2 + s], dict(c=normal_color(), linewidth=0.4)),
            ([0.9, 0.9 + 0.1 * s], [0.1, 0.1 + s], dict(c=normal_color(), linewidth=0.4)),
            ([1.0, 1.0 + 0.1 * s], [0.0, s], dict(c=normal_color(), linewidth=0.4)),
            # ticks of Caly
            ([0.0, s], [0.0, -s], dict(c=normal_color(), linewidth=0.4)),
            ([0.1, 0.1 + s], [0.0, -s], dict(c=normal_color(), linewidth=0.4)),
            ([0.2, 0.2 + s], [0.0, -s], dict(c=normal_color(), linewidth=0.4)),
            ([0.3, 0.3 + s], [0.0, -s], dict(c=normal_color(), linewidth=0.4)),
            ([0.4, 0.4 + s], [0.0, -s], dict(c=normal_color(), linewidth=0.4)),
            ([0.5, 0.5 + s], [0.0, -s], dict(c=normal_color(), linewidth=0.4)),
            ([0.6, 0.6 + s], [0.0, -s], dict(c=normal_color(), linewidth=0.4)),
            ([0.7, 0.7 + s], [0.0, -s], dict(c=normal_color(), linewidth=0.4)),
            ([0.8, 0.8 + s], [0.0, -s], dict(c=normal_color(), linewidth=0.4)),
            ([0.9, 0.9 + s], [0.0, -s], dict(c=normal_color(), linewidth=0.4)),
            ([1.0, 1.0 - s], [0.0, -s], dict(c=normal_color(), linewidth=0.4)),

            # guide lines
            ([0.03, 0.03 + 4 * s], [0.0, -2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.13, 0.13 + 4 * s], [0.0, -2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.33, 0.33 + 4 * s], [0.0, -2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.63, 0.63 + 4 * s], [0.0, -2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.83, 0.83 + 4 * s], [0.0, -2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.97, 0.97], [0.0, -2 * s], dict(c=normal_color(), linewidth=0.4)),

            ([0.0, -2 * s], [0.03, 0.03 + 2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2 * s], [0.13, 0.13 + 2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2 * s], [0.33, 0.33 + 2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2 * s], [0.63, 0.63 + 2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2 * s], [0.83, 0.83 + 2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.0, -2 * s], [0.97, 0.97], dict(c=normal_color(), linewidth=0.4)),

            ([0.97, 0.97 + 0.2 * s], [0.03, 0.03 + 2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.87, 0.87 + 0.2 * s], [0.13, 0.13 + 2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.67, 0.67 + 0.2 * s], [0.33, 0.33 + 2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.37, 0.37 + 0.2 * s], [0.63, 0.63 + 2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.17, 0.17 + 0.2 * s], [0.83, 0.83 + 2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.03, 0.03 + 0.2 * s], [0.97, 0.97 + 2 * s], dict(c=normal_color(), linewidth=0.4)),

            ([0.03, 0.03 + 2 * s], [0.95, 0.95 + 2 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.03, -2 * s], [0.03, 0.03 + 4 * s], dict(c=normal_color(), linewidth=0.4)),
            ([0.94, 0.94 + 0.4 * s], [0.03, 0.03 + 4 * s], dict(c=normal_color(), linewidth=0.4)),

        ]
        return STRUCTURAL_LINES + ADDITIONAL_LINES

    @property
    def labels(self):
        s = 0.01
        LABELS = [
            ((-s, -s), "C", dict(ha="right", va="top", fontsize=8, fontweight="bold")),
            ((-2 * s, 1 + 4 * s), "S", dict(ha="center", va="bottom", fontsize=8, fontweight="bold")),
            ((1 + 2 * s, -s), "SI", dict(ha="left", va="top", fontsize=8, fontweight="bold")),
            ((-5 * s, 0.55), "Sand %", dict(ha="right", va="center", fontsize=8, fontweight="bold")),
            ((0.5, 0.55), "Silt %", dict(ha="left", va="center", fontsize=8, fontweight="bold")),
            ((0.5 + 5 * s, -5 * np.sqrt(3) * s), "Clay %", dict(ha="center", va="top", fontsize=8, fontweight="bold")),
            # tick labels of Sand
            ((-s, 1.0 + s), "100%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.9 + s), "90%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.8 + s), "80%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.7 + s), "70%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.6 + s), "60%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.5 + s), "50%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.4 + s), "40%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.3 + s), "30%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.2 + s), "20%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.1 + s), "10%", dict(ha="right", va="bottom", fontsize=7)),
            ((-s, 0.0 + s), "0%", dict(ha="right", va="bottom", fontsize=7)),
            # tick labels of Silt
            ((0.0, 1.0 + s), "0%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.1, 0.9 + s), "10%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.2, 0.8 + s), "20%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.3, 0.7 + s), "30%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.4, 0.6 + s), "40%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.5, 0.5 + s), "50%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.6, 0.4 + s), "60%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.7, 0.3 + s), "70%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.8, 0.2 + s), "80%", dict(ha="left", va="bottom", fontsize=7)),
            ((0.9, 0.1 + s), "90%", dict(ha="left", va="bottom", fontsize=7)),
            ((1.0, 0.0 + s), "100%", dict(ha="left", va="bottom", fontsize=7)),
            # tick labels of Clay
            ((2 * s, -s), "100%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.1 + 2 * s, -s), "90%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.2 + 2 * s, -s), "80%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.3 + 2 * s, -s), "70%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.4 + 2 * s, -s), "60%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.5 + 2 * s, -s), "50%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.6 + 2 * s, -s), "40%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.7 + 2 * s, -s), "30%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.8 + 2 * s, -s), "20%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((0.9 + 2 * s, -s), "10%", dict(ha="center", va="top", fontsize=7, rotation=-45)),
            ((1.0 - s, -s), "0%", dict(ha="center", va="top", fontsize=7, rotation=-45)),

            ((0.125, 0.025), "(vs)(si)C", dict(ha="center", va="center", fontsize=6)),
            ((0.35, 0.025), "(vs)siC", dict(ha="center", va="center", fontsize=6)),
            ((0.625, 0.025), "(vs)cSI", dict(ha="center", va="center", fontsize=6)),
            ((0.850, 0.025), "(vs)(c)SI", dict(ha="center", va="center", fontsize=6)),

            ((0.125, 0.125), "(s)(si)C", dict(ha="center", va="center", fontsize=6)),
            ((0.3, 0.125), "(s)siC", dict(ha="center", va="center", fontsize=6)),
            ((0.575, 0.125), "(s)cSI", dict(ha="center", va="center", fontsize=6)),
            ((0.750, 0.125), "(s)(c)SI", dict(ha="center", va="center", fontsize=6)),
            ((0.125, 0.35), "(si)sC", dict(ha="center", va="center", fontsize=6)),
            ((0.3, 0.3), "ssiC", dict(ha="center", va="center", fontsize=6)),
            ((0.4, 0.3), "scSI", dict(ha="center", va="center", fontsize=6)),
            ((0.525, 0.35), "(c)sSI", dict(ha="center", va="center", fontsize=6)),
            ((0.125, 0.55), "(si)cS", dict(ha="center", va="center", fontsize=6)),
            ((1 / 3 - (0.45 - 1 / 3) / 2, 0.45), "sicS", dict(ha="center", va="center", fontsize=6)),
            ((0.325, 0.55), "(c)siS", dict(ha="center", va="center", fontsize=6)),
            ((0.125, 0.75), "(si)(c)S", dict(ha="center", va="center", fontsize=6)),

            ((0.03, 0.125), "(vsi)(s)C", dict(ha="center", va="center", fontsize=6, rotation=60)),
            ((0.03, 0.35), "(vsi)sC", dict(ha="center", va="center", fontsize=6, rotation=60)),
            ((0.03, 0.6), "(vsi)cS", dict(ha="center", va="center", fontsize=6, rotation=60)),
            ((0.03, 0.85), "(vsi)(c)S", dict(ha="center", va="center", fontsize=6, rotation=60)),

            ((0.845, 0.125), "(vc)(s)SI", dict(ha="center", va="center", fontsize=6, rotation=-60)),
            ((0.62, 0.35), "(vc)sSI", dict(ha="center", va="center", fontsize=6, rotation=-60)),
            ((0.37, 0.6), "(vc)siS", dict(ha="center", va="center", fontsize=6, rotation=-60)),
            ((0.12, 0.85), "(vc)(si)S", dict(ha="center", va="center", fontsize=6, rotation=-60)),

            ((0.05 + 4 * s, -2 * s), "(vsi)C", dict(ha="center", va="top", fontsize=6, rotation=-45)),
            ((0.14 + 4 * s, -2 * s), "(si)C", dict(ha="center", va="top", fontsize=6)),
            ((0.34 + 4 * s, -2 * s), "siC", dict(ha="center", va="top", fontsize=6)),
            ((0.64 + 4 * s, -2 * s), "cSI", dict(ha="center", va="top", fontsize=6)),
            ((0.84 + 4 * s, -2 * s), "(c)SI", dict(ha="center", va="top", fontsize=6)),
            ((0.94 + 4 * s, -2 * s), "(vc)SI", dict(ha="center", va="top", fontsize=6, rotation=-45)),

            ((-2 * s, 0.03 + 2 * s), "(vs)C", dict(ha="right", va="center", fontsize=6)),
            ((-2 * s, 0.13 + 2 * s), "(s)C", dict(ha="right", va="bottom", fontsize=6)),
            ((-2 * s, 0.33 + 2 * s), "sC", dict(ha="right", va="bottom", fontsize=6)),
            ((-2 * s, 0.63 + 2 * s), "cS", dict(ha="right", va="bottom", fontsize=6)),
            ((-2 * s, 0.83 + 2 * s), "(c)S", dict(ha="right", va="bottom", fontsize=6)),
            ((-2 * s, 0.97), "(vc)S", dict(ha="right", va="center", fontsize=6)),

            ((0.97 + 0.2 * s, 0.03 + 2 * s), "(vs)SI", dict(ha="left", va="center", fontsize=6)),
            ((0.87 + 0.2 * s, 0.13 + 2 * s), "(s)SI", dict(ha="left", va="bottom", fontsize=6)),
            ((0.67 + 0.2 * s, 0.33 + 2 * s), "sSI", dict(ha="left", va="bottom", fontsize=6)),
            ((0.37 + 0.2 * s, 0.63 + 2 * s), "siS", dict(ha="left", va="bottom", fontsize=6)),
            ((0.17 + 0.2 * s, 0.83 + 2 * s), "(si)S", dict(ha="left", va="bottom", fontsize=6)),
            ((0.03 + 0.2 * s, 0.97 + 2 * s), "(vsi)S", dict(ha="left", va="bottom", fontsize=6)),

            ((0.03 + 0.5 * s, 0.97 + 2 * s), "(vsi)(vc)S", dict(ha="left", va="top", fontsize=6)),
            ((-2 * s, 0.03 + 4 * s), "(vs)(vsi)C", dict(ha="right", va="bottom", fontsize=6)),
            ((0.94 + 0.4 * s, 0.03 + 4 * s), "(vs)(vc)SI", dict(ha="left", va="bottom", fontsize=6))]

        return LABELS

    def trans_pos(self, a, b):
        a, b = np.array(a), np.array(b)
        y = 0.5 * np.sqrt(3) * b
        x = 0.5 * b + a
        return x, y

    def plot_legend(self):
        x0, y0 = 0.7, 0.9
        w, h = 0.5, 0.16
        span = 0.01
        row_h = 0.02
        self.axes.plot([x0, x0 + w, x0 + w, x0, x0], [y0, y0, y0 - h, y0 - h, y0], c=normal_color(), linewidth=0.8)
        self.axes.text(x0 + w / 2, y0 - span, "Conventions", ha="center", va="top", fontsize=6,
                       color=normal_color())
        texts = ["UPPER CASE - Largest component (noun)",
                 "Lower case - Descriptive term (adjective)",
                 "( ) - Slightly (qualification)",
                 "(v ) - Very slightly (qualification)"]
        for row, text in enumerate(texts, 1):
            x = x0 + span
            y = y0 - span - row * (row_h + span)
            self.axes.text(x, y, text, ha="left", va="top", fontsize=6, color=normal_color())

        x0, y0 = 0.9, 0.7
        w, h = 0.3, 0.32
        span = 0.01
        row_h = 0.02
        self.axes.plot([x0, x0 + w, x0 + w, x0, x0], [y0, y0, y0 - h, y0 - h, y0], c=normal_color(), linewidth=0.8)
        self.axes.text(x0 + w / 2, y0 - span, "Term used", ha="center", va="top", fontsize=6,
                       color=normal_color())
        texts = ["S  - Sand     s  - Sandy",
                 "SI - Silt     si - Silty",
                 "C  - Clay     c  - Clayey",
                 "(s)   - Slightly sandy",
                 "(si)  - Slightly silty",
                 "(c)   - Slightly clayey",
                 "(vs)  - Very slightly sandy",
                 "(vsi) - Very slightly silty",
                 "(vc)  - Very slightly clayey"]
        for row, text in enumerate(texts, 1):
            x = x0 + span
            y = y0 - span - row * (row_h + span)
            self.axes.text(x, y, text, ha="left", va="top", fontsize=6, color=normal_color())

    def convert_samples(self, samples: List[Sample]) -> Tuple[Sequence[float], Sequence[float]]:
        silt = []
        sand = []
        for i, sample in enumerate(samples):
            sand_i, silt_i, clay_i = proportions_ssc(sample.classes_phi, sample.distribution)
            silt.append(silt_i)
            sand.append(sand_i)

        return silt, sand

    def retranslate(self):
        self.setWindowTitle(self.tr("SSC Diagram (Blott & Pye, 2012)"))
        self.edit_figure_action.setText(self.tr("Edit Figure"))
        self.save_figure_action.setText(self.tr("Save Figure"))


class CMDiagramChart(DiagramChart):
    """
    The C-M diagram referred to Mycielska-Dowgiałło and Ludwikowska-Kędzia (2011).

    Mycielska-Dowgiałło, E., Ludwikowska-Kędzia, M., 2011. Alternative interpretations of grain-size data from
        Quaternary deposits. Geologos 17. https://doi.org/10.2478/v10118-011-0010-9

    """
    GRID_C = (0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.4, 0.6, 0.8, 1.0, 2.0, 4.0, 6.0, 8.0, 10.0, 20.0, 40.0)
    GRID_M = (0.004, 0.01, 0.015, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.4, 0.6, 0.8)

    def __init__(self, parent=None, figsize=(4.4, 5.5)):
        super().__init__(parent=parent, figsize=figsize)
        self.setWindowTitle(self.tr("C-M Diagram"))

    @property
    def title(self) -> str:
        return "C-M Diagram"

    @property
    def lines(self):
        STRUCTURAL_LINES = [([0.003, 0.01, 1.0, 1.0, 0.003, 0.003], [0.01, 0.01, 1.0, 60.0, 60.0, 0.01],
                             dict(c=normal_color(), linewidth=0.8))]
        for m in self.GRID_M:
            line = ([m, m], [max(m, 0.01), 60.0], dict(c=normal_color(), linewidth=0.5))
            STRUCTURAL_LINES.append(line)
        for c in self.GRID_C:
            line = ([0.003, min(c, 1.0)], [c, c], dict(c=normal_color(), linewidth=0.5))
            STRUCTURAL_LINES.append(line)

        ADDITIONAL_LINES = [([0.365, 1.25], [0.600, 0.600], dict(c=normal_color(), linestyle="--", linewidth=0.8)),
                            ([0.50, 1.25], [0.900, 0.900], dict(c=normal_color(), linestyle="--", linewidth=0.8)),
                            ([0.50, 1.25], [18.00, 18.00], dict(c=normal_color(), linestyle="--", linewidth=0.8)),
                            ([0.15, 1.25], [0.2383, 0.2383], dict(c=normal_color(), linestyle="--", linewidth=0.8)),
                            ([0.27, 1.25], [0.4366, 0.4366], dict(c=normal_color(), linestyle="--", linewidth=0.8)),
                            ([0.50, 1.25], [1.0734, 1.0734], dict(c=normal_color(), linestyle="--", linewidth=0.8)),

                            ([0.7, 0.7], [2.05, 1.5], dict(c=normal_color(), linestyle="-", linewidth=0.5)),
                            ([0.24, 0.275], [2.00, 1.2], dict(c=normal_color(), linestyle="-", linewidth=0.5)),
                            ([0.15, 0.2], [1.0, 0.9], dict(c=normal_color(), linestyle="-", linewidth=0.5)),
                            ([0.35, 0.15], [0.32, 0.32], dict(c=normal_color(), linestyle="-", linewidth=0.5)),
                            ([0.07, 0.07], [0.16, 0.2], dict(c=normal_color(), linestyle="-", linewidth=0.5)),

                            ([0.006, 0.006], [0.065, 0.05], dict(c=normal_color(), linestyle="-", linewidth=0.5)),
                            ([0.5, 0.5], [40, 25], dict(c=normal_color(), linestyle="-", linewidth=0.5)),
                            ([0.25, 0.35], [5.0, 5.0], dict(c=normal_color(), linestyle="-", linewidth=0.5)),
                            ([0.58, 0.30], [0.5, 0.8], dict(c=normal_color(), linestyle="-", linewidth=0.5)),
                            ([0.02, 0.025], [0.95, 0.5], dict(c=normal_color(), linestyle="-", linewidth=0.5)),
                            ([0.01, 0.02], [0.15, 0.1], dict(c=normal_color(), linestyle="-", linewidth=0.5))]
        for x, y, kwargs in ADDITIONAL_LINES:
            kwargs["zorder"] = 10

        return STRUCTURAL_LINES + ADDITIONAL_LINES

    @property
    def labels(self):
        s = 0.05
        LABELS = [
            ((0.00125, 2.0), "C (first percentile)",
             dict(ha="right", va="center", rotation=90.0, fontsize=8, fontweight="bold")),
            ((0.05, 110.0), "M (median)", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((0.15, 0.15), "C = M", dict(ha="center", va="top", rotation=45.0, fontsize=8, fontweight="bold")),
            ((0.003 * (1 - s), 60.0), "mm", dict(ha="right", va="center", fontsize=6)),
            ((0.002 * (1 - s), 60.0), r"phi ($\phi$)", dict(ha="right", va="center", fontsize=6)),
            ((1.0, 60.0 * (1 + s)), "1.0", dict(ha="center", va="bottom", fontsize=6)),
            ((1.0, 70.0 * (1 + s)), "0.0", dict(ha="center", va="bottom", fontsize=6)),
            ((1.5, 60.0 * (1 + s)), "mm", dict(ha="center", va="bottom", fontsize=6)),
            ((1.5, 70.0 * (1 + s)), r"phi ($\phi$)", dict(ha="center", va="bottom", fontsize=6)),

            ((0.70, 3.30), "I", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((0.15, 3.30), "II", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((0.03, 3.30), "III", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((0.51, 0.68), "IV", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((0.15, 0.68), "V", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((0.05, 0.13), "VI", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((0.03, 0.68), "VII", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((0.007, 0.68), "VIII", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((0.007, 3.30), "IX", dict(ha="center", va="center", fontsize=8, fontweight="bold")),

            ((0.0175, 0.200), "S", dict(ha="center", va="center", fontsize=7)),
            ((0.1168, 0.234), "R", dict(ha="center", va="center", fontsize=7)),
            ((0.2077, 0.436), "Q", dict(ha="center", va="center", fontsize=7)),
            ((0.2200, 1.142), "P", dict(ha="center", va="center", fontsize=7)),
            ((0.5500, 1.265), "O", dict(ha="center", va="center", fontsize=7)),
            ((0.8000, 1.482), "N", dict(ha="center", va="center", fontsize=7)),
            ((0.0055, 0.032), "T", dict(ha="center", va="center", fontsize=7)),

            ((0.0125, 0.475), "S", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((0.1500, 0.475), "R", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((0.3500, 0.975), "Q", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((0.3700, 23.700), "P", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((0.7000, 23.700), "O", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((0.0250, 0.125), "T", dict(ha="center", va="center", fontsize=8, fontweight="bold")),

            ((1.5, 0.595), "Cu", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((1.5, 0.900), "Cs", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((1.5, 18.00), "Cr", dict(ha="center", va="center", fontsize=8, fontweight="bold")),
            ((1.5, 0.2383), "Cu", dict(ha="center", va="center", fontsize=8, fontstyle="italic")),
            ((1.5, 0.4366), "Cs", dict(ha="center", va="center", fontsize=8, fontstyle="italic")),
            ((1.5, 1.0734), "Cr", dict(ha="center", va="center", fontsize=8, fontstyle="italic")),

            ((0.7, 2.25), "rolling", dict(ha="center", va="center", fontsize=6)),
            ((0.15, 2.0), "rolling and\nsuspension",
             dict(ha="center", va="center", fontsize=6, backgroundcolor=background_color())),
            ((0.10, 1.2), "suspension\nand rolling",
             dict(ha="center", va="center", fontsize=6, backgroundcolor=background_color())),
            ((0.36, 0.32), "graded suspension", dict(ha="left", va="center", fontsize=6)),
            ((0.075, 0.125), "uniform\nsuspension", dict(ha="center", va="center", fontsize=6)),
            ((0.0068, 0.08), "pelagic\nsuspension",
             dict(ha="center", va="center", fontsize=6, backgroundcolor=background_color())),

            ((0.5, 45.0), "rolling", dict(ha="center", va="center", fontsize=6, backgroundcolor=background_color())),
            ((0.25, 4.86), "saltation and rolling",
             dict(ha="right", va="center", fontsize=6, backgroundcolor=background_color())),
            ((0.60, 0.55), "saltation", dict(ha="left", va="top", fontsize=6)),
            ((0.02, 1.0), "suspension", dict(ha="center", va="bottom", fontsize=6, backgroundcolor=background_color())),
            ((0.0068, 0.14), "overbank-pool\nfacies\nsuspension",
             dict(ha="center", va="center", fontsize=6, backgroundcolor=background_color()))]

        for m in self.GRID_M:
            label_mm = ((m, 60.0 * (1 + s)), f"{m}"[1:], dict(ha="center", va="bottom", fontsize=6))
            label_phi = ((m, 70.0 * (1 + s)), f"{to_phi(m * 1000):0.1f}", dict(ha="center", va="bottom", fontsize=6))
            LABELS.append(label_mm)
            LABELS.append(label_phi)
        for c in self.GRID_C:
            label_mm = ((0.003 * (1 - s), c), f"{c}", dict(ha="right", va="center", fontsize=6))
            label_phi = ((0.002 * (1 - s), c), f"{to_phi(c * 1000):0.1f}", dict(ha="right", va="center", fontsize=6))
            LABELS.append(label_mm)
            LABELS.append(label_phi)

        return LABELS

    def trans_pos(self, a, b):
        return np.log10(a), np.log10(b)

    def plot_legend(self):
        def pixel_to_mc(x, y):
            m = 0.01 * 10 ** ((x - 388) / 400)
            c = 20.0 / 10 ** ((y - 345) / 407)
            return m, c

        passega_t = Circle(self.trans_pos(*pixel_to_mc(284, 1484)), radius=(369 - 284) / 400,
                           linewidth=1.0, edgecolor=normal_color(), facecolor="#747474", alpha=0.8)
        self.axes.add_patch(passega_t)
        passega_chanel = dict(data=[
            (Path.MOVETO, (512, 1182)), (Path.LINETO, (781, 1182)), (Path.CURVE3, (802, 1178)),
            (Path.CURVE3, (824, 1159)), (Path.LINETO, (925, 1059)), (Path.CURVE3, (938, 1039)),
            (Path.CURVE3, (940, 1027)), (Path.LINETO, (940, 892)), (Path.CURVE4, (944, 873)),
            (Path.CURVE4, (952, 861)), (Path.CURVE4, (979, 859)), (Path.LINETO, (1092, 859)),
            (Path.CURVE3, (1126, 852)), (Path.CURVE3, (1145, 838)), (Path.LINETO, (1157, 829)),
            (Path.CURVE4, (1198, 786)), (Path.CURVE4, (1165, 747)), (Path.CURVE4, (1120, 793)),
            (Path.CURVE3, (1105, 805)), (Path.CURVE3, (1073, 811)), (Path.LINETO, (948, 811)),
            (Path.CURVE4, (889, 811)), (Path.CURVE4, (889, 863)), (Path.CURVE4, (889, 877)),
            (Path.LINETO, (891, 1007)), (Path.CURVE3, (885, 1029)), (Path.CURVE3, (870, 1043)),
            (Path.LINETO, (806, 1106)), (Path.CURVE3, (781, 1127)), (Path.CURVE3, (751, 1133)),
            (Path.LINETO, (481, 1133)), (Path.CURVE4, (426, 1133)), (Path.CURVE4, (426, 1182)),
            (Path.CURVE4, (481, 1182)), (Path.CLOSEPOLY, (512, 1182))],
            linewidth=1.0, facecolor="#747474", alpha=0.8)
        lk2000_chanel = dict(data=[
            (Path.MOVETO, (512, 1094)), (Path.LINETO, (817, 1094)), (Path.CURVE3, (890, 1094)),
            (Path.CURVE3, (925, 1056)), (Path.LINETO, (1037, 946)), (Path.CURVE3, (1061, 911)),
            (Path.CURVE3, (1067, 859)), (Path.LINETO, (1067, 395)), (Path.CURVE3, (1065, 354)),
            (Path.CURVE3, (1110, 350)), (Path.CURVE4, (1165, 350)), (Path.CURVE4, (1165, 272)),
            (Path.CURVE4, (1110, 272)), (Path.LINETO, (1037, 272)), (Path.CURVE4, (1000, 272)),
            (Path.CURVE4, (985, 287)), (Path.CURVE4, (985, 324)), (Path.LINETO, (985, 821)),
            (Path.CURVE3, (985, 862)), (Path.CURVE3, (964, 893)), (Path.LINETO, (907, 945)),
            (Path.CURVE3, (882, 965)), (Path.CURVE3, (839, 965)), (Path.LINETO, (456, 965)),
            (Path.CURVE4, (363, 965)), (Path.CURVE4, (363, 1094)), (Path.CURVE4, (456, 1094)),
            (Path.CLOSEPOLY, (512, 1094))],
            linewidth=1.0, facecolor="#cfcfcf", alpha=0.8)
        lk2000_t = dict(data=[
            (Path.MOVETO, (512, 1339)), (Path.LINETO, (669, 1183)), (Path.CURVE4, (709, 1143)),
            (Path.CURVE4, (658, 1091)), (Path.CURVE4, (618, 1131)), (Path.LINETO, (444, 1304)),
            (Path.CURVE4, (404, 1344)), (Path.CURVE4, (456, 1397)), (Path.CURVE4, (496, 1357)),
            (Path.CLOSEPOLY, (512, 1339))],
            linewidth=1.0, facecolor="#cfcfcf", alpha=0.8)

        patchs = [lk2000_chanel, lk2000_t, passega_chanel]
        for patch_define in patchs:
            codes, pixels = zip(*patch_define["data"])
            verts = []
            for x, y in pixels:
                verts.append(self.trans_pos(*pixel_to_mc(x, y)))
            path = Path(verts, codes)
            patch = PathPatch(path, linewidth=patch_define["linewidth"], edgecolor=normal_color(),
                              facecolor=patch_define["facecolor"], alpha=patch_define["alpha"])
            self.axes.add_patch(patch)
        # caption = "This C-M diagram is modified after Mycielska-Dowgiałło & Ludwikowska-Kędzia (2011). " \
        #           "See the description of symbols from Passega (1964),
        #           "Passega & Byramjee (1969) and Mycielska-Dowgiałło & Ludwikowska-Kędzia (2011)."
        # self.axes.text(*self.trans_pos(0.003, 0.01), caption, fontsize=6, ha="left", va="top", wrap=True)

    def convert_samples(self, samples: List[Sample]) -> Tuple[Sequence[float], Sequence[float]]:
        c = []
        m = []
        for i, sample in enumerate(samples):
            c_i, m_i = cm(sample.classes_phi, sample.distribution)
            c.append(to_microns(c_i) / 1000)
            m.append(to_microns(m_i) / 1000)
        return m, c

    def retranslate(self):
        self.setWindowTitle(self.tr("C-M Diagram"))
        self.edit_figure_action.setText(self.tr("Edit Figure"))
        self.save_figure_action.setText(self.tr("Save Figure"))



import logging
import os
import socket

QGRAIN_VERSION = "0.5.3.4"
QGRAIN_ROOT_PATH = os.path.dirname(__file__)

HELLO_TEXT = r"""
 _______  _______  _______  _______ _________ _       
(  ___  )(  ____ \(  ____ )(  ___  )\__   __/( (    /|
| (   ) || (    \/| (    )|| (   ) |   ) (   |  \  ( |
| |   | || |      | (____)|| (___) |   | |   |   \ | |
| |   | || | ____ |     __)|  ___  |   | |   | (\ \) |
| | /\| || | \_  )| (\ (   | (   ) |   | |   | | \   |
| (_\ \ || (___) || ) \ \__| )   ( |___) (___| )  \  |
(____\/_)(_______)|/   \__/|/     \|\_______/|/    )_)

An easy-to-use software for the analysis of grain size distributions

"""


def get_free_tcp_port():
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind(("", 0))
    _, port = tcp.getsockname()
    tcp.close()
    return port


def main():
    print(HELLO_TEXT)
    import argparse
    import numpy as np
    np.seterr(all="ignore")
    parser = argparse.ArgumentParser(
        description="QGrain is an easy-to-use software for the analysis of grain size distributions.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-s", "--server", action="store_true", default=False,
                        help="start a local grpc server to handle the requests of computation")
    parser.add_argument("--address", type=str, default="localhost:50051",
                        help="specify the local ip address used to start the grpc server")
    parser.add_argument("--max-workers", type=int, default=4, help="specify the max number of thread workers.",
                        dest="max_workers")
    parser.add_argument("--max-message-length", type=int, default=2**30,
                        help="specify the max length of grpc messages (bytes).", dest="max_message_length")
    parser.add_argument("--max-dataset-size", type=int, default=100000, help="specify the max size of dataset.",
                        dest="max_dataset_size")
    parser.add_argument("--target", type=str, default="",
                        help="specify the remote ip address of the grpc server")
    args = parser.parse_args()
    if args.server:
        from .protos.server import QGrainServicer
        from .ui import create_necessary_folders
        create_necessary_folders()
        format_str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        from logging.handlers import TimedRotatingFileHandler
        file_handler = TimedRotatingFileHandler(
            os.path.join(os.path.expanduser("~"), "QGrain", "logs", "qgrain_server.log"),
            when="D", backupCount=8, encoding="utf-8")
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(logging.Formatter(format_str))
        logging.basicConfig(level=logging.DEBUG, format=format_str)
        logging.getLogger().addHandler(file_handler)
        qgrain_server = QGrainServicer(
            address=args.address, max_workers=args.max_workers, max_message_length=args.max_message_length,
            max_dataset_size=args.max_dataset_size)
        qgrain_server.serve()
    else:
        from PySide6 import QtCore
        from .protos.client import QGrainClient
        from .ui import setup_app, setup_logging, create_necessary_folders
        from .ui.MainWindow import MainWindow
        create_necessary_folders()
        if args.target != "":
            QGrainClient.set_target(args.target)
        settings = QtCore.QSettings(os.path.join(os.path.expanduser("~"), "QGrain", "qgrain.ini"),
                                    QtCore.QSettings.Format.IniFormat)
        language = settings.value("language", "en", type=str)
        theme = settings.value("theme", "default", type=str)
        app = setup_app(language, theme)
        main_window = MainWindow()
        # load the artificial dataset to show the functions of all modules
        dataset = main_window.dataset_generator.get_random_dataset(100)
        main_window.load_dataset_dialog.dataset_loaded.emit(dataset.dataset)
        main_window.ssu_analyzer.on_try_fit_clicked()
        main_window.parameter_editor.refer_parameters(dataset.distribution_type, dataset.parameters[0])
        main_window.parameter_editor.enabled_checkbox.setChecked(True)
        main_window.emma_analyzer.on_try_fit_clicked()
        main_window.udm_analyzer.on_try_fit_clicked()
        main_window.parameter_editor.enabled_checkbox.setChecked(False)
        setup_logging(main_window.statusBar(), main_window.log_dialog)
        main_window.show()
        app.exec()



__all__ = ["DatasetGenerator"]

import logging
from typing import *

from PySide6 import QtCore, QtWidgets

from ..models import DistributionType
from ..generate import SIMPLE_PRESET, random_dataset, random_mean_sample, random_sample
from ..io import save_artificial_dataset
from ..charts.DistributionChart import DistributionChart


class GeneratorComponent(QtWidgets.QWidget):
    NORMAL_SETTINGS = dict(
        n_parameters=3,
        parameter_names=("Location", "Scale", "Weight"),
        mean_ranges=((-15.0, 15.0), (0.01, 100.0), (0.0, 10.0)),
        mean_defaults=(5.0, 1.0, 1.0),
        mean_steps=(0.1, 0.1, 0.1),
        std_ranges=((0.0, 10.0), (0.0, 10.0), (0.0, 10.0)),
        std_defaults=(0.0, 0.0, 0.1),
        std_steps=(0.1, 0.1, 0.1))

    SKEW_NORMAL_SETTINGS = dict(
        n_parameters=4,
        parameter_names=("Shape", "Location", "Scale", "Weight"),
        mean_ranges=((-100.0, 100.0), (-15.0, 15.0), (0.01, 100.0), (0.0, 10.0)),
        mean_defaults=(0.0, 5.0, 1.0, 1.0),
        mean_steps=(0.1, 0.1, 0.1, 0.1),
        std_ranges=((0.0, 10.0), (0.0, 10.0), (0.0, 10.0), (0.0, 10.0)),
        std_defaults=(0.0, 0.0, 0.0, 0.1),
        std_steps=(0.1, 0.1, 0.1, 0.1))

    WEIBULL_SETTINGS = dict(
        n_parameters=3,
        parameter_names=("Shape", "Scale", "Weight"),
        mean_ranges=((-500.0, 500.0), (0.01, 500.0), (0.0, 10.0)),
        mean_defaults=(3.6, 1.0, 1.0),
        mean_steps=(0.1, 0.1, 0.1),
        std_ranges=((0.0, 100.0), (0.0, 100.0), (0.0, 10.0)),
        std_defaults=(0.0, 0.0, 0.1),
        std_steps=(0.1, 0.1, 0.1))

    GENERAL_WEIBULL_SETTINGS = dict(
        n_parameters=4,
        parameter_names=("Shape", "Location", "Scale", "Weight"),
        mean_ranges=((-500.0, 500.0), (-500.0, 500.0), (0.01, 500.0), (0.0, 10.0)),
        mean_defaults=(3.6, 5.0, 1.0, 1.0),
        mean_steps=(0.1, 0.1, 0.1, 0.1),
        std_ranges=((0.0, 100.0), (0.0, 100.0), (0.0, 100.0), (0.0, 10.0)),
        std_defaults=(0.0, 0.0, 0.0, 0.1),
        std_steps=(0.1, 0.1, 0.1, 0.1))

    SETTING_MAP = {
        DistributionType.Normal: NORMAL_SETTINGS,
        DistributionType.SkewNormal: SKEW_NORMAL_SETTINGS,
        DistributionType.Weibull: WEIBULL_SETTINGS,
        DistributionType.GeneralWeibull: GENERAL_WEIBULL_SETTINGS}

    value_changed = QtCore.Signal()

    def __init__(self, name: str, distribution_type: DistributionType, parent=None):
        super().__init__(parent=parent)
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        # make sure that the pyside2-lupdate.exe can recognize these keywords
        _ = [self.tr("Shape"), self.tr("Location"), self.tr("Scale"), self.tr("Weight")]
        self.main_layout = QtWidgets.QGridLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.group = QtWidgets.QGroupBox("")
        self.group.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        self.main_layout.addWidget(self.group)
        self.group_layout = QtWidgets.QGridLayout(self.group)
        self.group_layout.setColumnMinimumWidth(0, 32)
        self.group_layout.setColumnMinimumWidth(1, 32)
        self.group_layout.setColumnMinimumWidth(2, 32)
        settings = self.SETTING_MAP[distribution_type]
        self.widgets = []  # type: list[tuple[QtWidgets.QLabel, QtWidgets.QDoubleSpinBox, QtWidgets.QDoubleSpinBox]]
        self.name_label = QtWidgets.QLabel(name)
        self.name_label.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
        self.mean_label = QtWidgets.QLabel(self.tr("Mean"))
        self.mean_label.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
        self.std_label = QtWidgets.QLabel(self.tr("Standard\nDeviation"))
        self.std_label.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
        self.group_layout.addWidget(self.name_label, 0, 0)
        self.group_layout.addWidget(self.mean_label, 0, 1)
        self.group_layout.addWidget(self.std_label, 0, 2)
        for i in range(settings["n_parameters"]):
            label = QtWidgets.QLabel(self.tr(settings["parameter_names"][i]))
            label.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
            mean_input = QtWidgets.QDoubleSpinBox()
            mean_input.setRange(*settings["mean_ranges"][i])
            mean_input.setSingleStep(settings["mean_steps"][i])
            mean_input.setValue(settings["mean_defaults"][i])
            std_input = QtWidgets.QDoubleSpinBox()
            std_input.setRange(*settings["std_ranges"][i])
            std_input.setSingleStep(settings["std_steps"][i])
            std_input.setValue(settings["std_defaults"][i])
            self.group_layout.addWidget(label, i + 1, 0)
            self.group_layout.addWidget(mean_input, i + 1, 1)
            self.group_layout.addWidget(std_input, i + 1, 2)
            self.widgets.append((label, mean_input, std_input))
        for _, mean_input, std_input in self.widgets:
            mean_input.valueChanged.connect(self.on_value_changed)
            std_input.valueChanged.connect(self.on_value_changed)
        self._name = name
        self._distribution_type = distribution_type

    @property
    def target(self):
        target = [(mean_input.value(), std_input.value()) for _, mean_input, std_input in self.widgets]
        return target

    @target.setter
    def target(self, values: list):
        for (mean, std), (_, mean_input, std_input) in zip(values, self.widgets):
            mean_input.setValue(mean)
            std_input.setValue(std)

    def on_value_changed(self, _):
        self.value_changed.emit()

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.mean_label.setText(self.tr("Mean"))
        self.std_label.setText(self.tr("Standard\nDeviation"))
        settings = self.SETTING_MAP[self._distribution_type]
        for param_name, (param_label, _, _) in zip(settings["parameter_names"], self.widgets):
            param_label.setText(self.tr(param_name))


class DatasetGenerator(QtWidgets.QWidget):
    logger = logging.getLogger("QGrain.DatasetGenerator")

    SUPPORT_DISTRIBUTIONS = (
        (DistributionType.Normal, "Normal"),
        (DistributionType.SkewNormal, "Skew Normal"),
        (DistributionType.Weibull, "Weibull"),
        (DistributionType.GeneralWeibull, "General Weibull"))

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)
        self.main_layout = QtWidgets.QGridLayout(self)
        self.control_group = QtWidgets.QGroupBox(self.tr("Control"))
        self.control_layout = QtWidgets.QGridLayout(self.control_group)
        self.control_layout.setColumnStretch(0, 1)
        self.control_layout.setColumnStretch(1, 1)
        self.n_components_label = QtWidgets.QLabel(self.tr("Number of Components"))
        self.n_components_input = QtWidgets.QSpinBox()
        self.n_components_input.setRange(1, 12)
        self.n_components_input.valueChanged.connect(self.on_n_components_changed)
        self.distribution_type_label = QtWidgets.QLabel(self.tr("Distribution Type"))
        self.distribution_type_combo_box = QtWidgets.QComboBox()
        self.distribution_type_combo_box.addItems([name for _, name in self.SUPPORT_DISTRIBUTIONS])
        self.distribution_type_combo_box.currentIndexChanged.connect(self.on_distribution_type_changed)
        self.control_layout.addWidget(self.n_components_label, 0, 0)
        self.control_layout.addWidget(self.n_components_input, 0, 1)
        self.control_layout.addWidget(self.distribution_type_label, 0, 2)
        self.control_layout.addWidget(self.distribution_type_combo_box, 0, 3)
        self.minimum_size_label = QtWidgets.QLabel(self.tr("Minimum Size"))
        self.minimum_size_input = QtWidgets.QDoubleSpinBox()
        self.minimum_size_input.setDecimals(2)
        self.minimum_size_input.setRange(1e-4, 1e6)
        self.minimum_size_input.setValue(0.0200)
        self.maximum_size_label = QtWidgets.QLabel(self.tr("Maximum Size"))
        self.maximum_size_input = QtWidgets.QDoubleSpinBox()
        self.maximum_size_input.setDecimals(2)
        self.maximum_size_input.setRange(1e-4, 1e6)
        self.maximum_size_input.setValue(2000.0000)
        self.control_layout.addWidget(self.minimum_size_label, 1, 0)
        self.control_layout.addWidget(self.minimum_size_input, 1, 1)
        self.control_layout.addWidget(self.maximum_size_label, 1, 2)
        self.control_layout.addWidget(self.maximum_size_input, 1, 3)
        self.n_classes_label = QtWidgets.QLabel(self.tr("Number of Classes"))
        self.n_classes_input = QtWidgets.QSpinBox()
        self.n_classes_input.setRange(10, 10000)
        self.n_classes_input.setValue(101)
        self.precision_label = QtWidgets.QLabel(self.tr("Data Precision"))
        self.precision_input = QtWidgets.QSpinBox()
        self.precision_input.setRange(2, 8)
        self.precision_input.setValue(4)
        self.control_layout.addWidget(self.n_classes_label, 2, 0)
        self.control_layout.addWidget(self.n_classes_input, 2, 1)
        self.control_layout.addWidget(self.precision_label, 2, 2)
        self.control_layout.addWidget(self.precision_input, 2, 3)
        self.n_samples_label = QtWidgets.QLabel(self.tr("Number of Samples"))
        self.n_samples_input = QtWidgets.QSpinBox()
        self.n_samples_input.setRange(100, 100000)
        self.control_layout.addWidget(self.n_samples_label, 3, 0)
        self.control_layout.addWidget(self.n_samples_input, 3, 1)
        self.preview_button = QtWidgets.QPushButton(self.tr("Preview"))
        self.preview_button.clicked.connect(self.on_preview_clicked)
        self.control_layout.addWidget(self.preview_button, 3, 2, 1, 2)
        self.parameter_tab_widget = QtWidgets.QTabWidget()
        self.component_sets: List[Tuple[QtWidgets.QWidget, QtWidgets.QGridLayout, List[GeneratorComponent]]] = []
        self._add_components()
        self.preview_group = QtWidgets.QGroupBox(self.tr("Preview"))
        self.preview_layout = QtWidgets.QGridLayout(self.preview_group)
        self.preview_layout.setContentsMargins(0, 0, 0, 0)
        self.chart = DistributionChart(parent=self)
        self.preview_layout.addWidget(self.chart, 0, 0)
        self.splitter_1 = QtWidgets.QSplitter()
        self.splitter_1.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_1.addWidget(self.control_group)
        self.splitter_1.addWidget(self.preview_group)
        # self.splitter_1.setStretchFactor(0, 2)
        # self.splitter_1.setStretchFactor(1, 1)
        self.splitter_2 = QtWidgets.QSplitter()
        self.splitter_2.setOrientation(QtCore.Qt.Vertical)
        self.splitter_2.addWidget(self.splitter_1)
        self.splitter_2.addWidget(self.parameter_tab_widget)
        self.main_layout.addWidget(self.splitter_2, 0, 0)

        self.distribution_type_combo_box.setCurrentIndex(1)
        self.target = SIMPLE_PRESET["target"]
        self.minimum_size_input.setValue(0.02)
        self.maximum_size_input.setValue(2000.0)
        self.n_classes_input.setValue(101)
        self.precision_input.setValue(4)
        self.file_dialog = QtWidgets.QFileDialog(parent=self)
        self.normal_msg = QtWidgets.QMessageBox(self)
        self.update_timer = QtCore.QTimer()
        self.update_timer.timeout.connect(lambda: self.update_chart(True))

    @property
    def n_components(self) -> int:
        return self.n_components_input.value()

    @property
    def distribution_type(self) -> DistributionType:
        index = self.distribution_type_combo_box.currentIndex()
        distribution_type = self.SUPPORT_DISTRIBUTIONS[index][0]
        return distribution_type

    @property
    def components(self) -> List[GeneratorComponent]:
        enabled = []
        for _, _, components in self.component_sets:
            for component in components:
                if component.isEnabled():
                    enabled.append(component)
        return enabled

    @property
    def target(self):
        return [component.target for component in self.components]

    @target.setter
    def target(self, values):
        if len(values) != len(self.components):
            self.n_components_input.setValue(len(values))
        for component_widget, component_target in zip(self.components, values):
            component_widget.blockSignals(True)
            component_widget.target = component_target
            component_widget.blockSignals(False)
        self.update_chart()

    @property
    def generate_kwargs(self):
        min_size = self.minimum_size_input.value()
        max_size = self.maximum_size_input.value()
        n_classes = self.n_classes_input.value()
        if min_size > max_size:
            min_size, max_size = max_size, min_size
        precision = self.precision_input.value()
        noise = precision + 1
        kwargs = dict(target=self.target, distribution_type=self.distribution_type,
                      min_size=min_size, max_size=max_size, n_classes=n_classes,
                      precision=precision, noise=noise)
        return kwargs

    def show_message(self, title: str, message: str):
        self.normal_msg.setWindowTitle(title)
        self.normal_msg.setText(message)
        self.normal_msg.exec_()

    def show_info(self, message: str):
        self.show_message(self.tr("Info"), message)

    def show_warning(self, message: str):
        self.show_message(self.tr("Warning"), message)

    def show_error(self, message: str):
        self.show_message(self.tr("Error"), message)

    def _clear_components(self):
        for i, (param_holder, param_layout, components) in enumerate(self.component_sets):
            for j, component in enumerate(components):
                component.value_changed.disconnect(self.on_value_changed)
                param_layout.removeWidget(component)
                component.hide()
            param_holder.hide()
        self.parameter_tab_widget.clear()
        self.component_sets.clear()

    def _add_components(self):
        for i in range(4):
            param_holder = QtWidgets.QWidget()
            self.parameter_tab_widget.addTab(param_holder, f"AC{i * 3 + 1}-{i * 3 + 3}")
            param_layout = QtWidgets.QGridLayout(param_holder)
            param_layout.setContentsMargins(0, 0, 0, 0)
            components = []
            for j in range(3):
                component = GeneratorComponent(f"AC{i * 3 + j + 1}", self.distribution_type)
                component.value_changed.connect(self.on_value_changed)
                param_layout.addWidget(component, 0, j)
                components.append(component)
            self.component_sets.append((param_holder, param_layout, components))

    def on_n_components_changed(self, n_components: int):
        for i in range(4):
            for j in range(3):
                current_index = i * 3 + j
                if current_index < n_components:
                    self.component_sets[i][2][j].setEnabled(True)
                else:
                    self.component_sets[i][2][j].setEnabled(False)

    def on_distribution_type_changed(self, _):
        self._clear_components()
        self._add_components()
        self.on_n_components_changed(self.n_components)

    def on_preview_clicked(self):
        if self.update_timer.isActive():
            self.preview_button.setText(self.tr("Preview"))
            self.update_timer.stop()
            self.update_chart()
        else:
            self.preview_button.setText(self.tr("Stop"))
            self.update_timer.start(500)

    def on_save_clicked(self):
        filename, _ = self.file_dialog.getSaveFileName(
            self, self.tr("Choose a filename to save the generated dataset"),
            ".", "Microsoft Excel (*.xlsx)")
        if filename is None or filename == "":
            return
        self.logger.debug(f"Generating a random dataset with the parameters: {self.generate_kwargs}.")
        n_samples = self.n_samples_input.value()
        dataset = self.get_random_dataset(n_samples)
        progress_dialog = QtWidgets.QProgressDialog(
            self.tr("Saving the generated dataset..."), self.tr("Cancel"),
            0, 100, self)
        progress_dialog.setWindowTitle("QGrain")
        progress_dialog.setWindowModality(QtCore.Qt.WindowModal)

        def callback(progress: float):
            if progress_dialog.wasCanceled():
                raise StopIteration()
            progress_dialog.setValue(int(progress * 100))
            QtCore.QCoreApplication.processEvents()
        try:
            save_artificial_dataset(dataset, filename, progress_callback=callback, logger=self.logger)
        except StopIteration:
            self.logger.info("The saving task was canceled.")
        finally:
            progress_dialog.close()

    def get_random_sample(self):
        if self.minimum_size_input.value() == self.maximum_size_input.value():
            return
        sample = random_sample(**self.generate_kwargs)
        sample.name = "Artificial Sample"
        return sample

    def get_mean_sample(self):
        if self.minimum_size_input.value() == self.maximum_size_input.value():
            return
        sample = random_mean_sample(**self.generate_kwargs)
        sample.name = "Artificial Sample"
        return sample

    def get_random_dataset(self, n_samples: int):
        if self.minimum_size_input.value() == self.maximum_size_input.value():
            return
        dataset = random_dataset(**self.generate_kwargs, n_samples=n_samples)
        return dataset

    def on_value_changed(self):
        self.update_chart()

    def update_chart(self, random=False):
        if not random:
            sample = self.get_mean_sample()
        else:
            sample = self.get_random_sample()
        self.chart.show_chart(sample)

    def changeEvent(self, event: QtCore.QEvent):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslate()

    def retranslate(self):
        self.control_group.setTitle(self.tr("Control"))
        self.preview_group.setTitle(self.tr("Preview"))
        self.n_components_label.setText(self.tr("Number of Components"))
        self.distribution_type_label.setText(self.tr("Distribution Type"))
        self.minimum_size_label.setText(self.tr("Minimum Size"))
        self.maximum_size_label.setText(self.tr("Maximum Size"))
        self.n_classes_label.setText(self.tr("Number of Classes"))
        self.precision_label.setText(self.tr("Data Precision"))
        self.n_samples_label.setText(self.tr("Number of Samples"))
        if self.update_timer.isActive():
            self.preview_button.setText(self.tr("Stop"))
        else:
            self.preview_button.setText(self.tr("Preview"))
